"""> PyDRex: 2D simple shear tests."""

import contextlib as cl
import functools as ft
import sys
from time import process_time

import numpy as np
import pytest
from numpy import asarray as Ŋ
from numpy import testing as nt
from scipy.interpolate import PchipInterpolator

from pydrex import core as _core
from pydrex import diagnostics as _diagnostics
from pydrex import geometry as _geo
from pydrex import io as _io
from pydrex import logger as _log
from pydrex import minerals as _minerals
from pydrex import pathlines as _paths
from pydrex import stats as _stats
from pydrex import utils as _utils
from pydrex import velocity as _velocity
from pydrex import visualisation as _vis

Pool, HAS_RAY = _utils.import_proc_pool()

# Subdirectory of `outdir` used to store outputs from these tests.
SUBDIR = "2d_simple_shear"


class TestPreliminaries:
    """Preliminary tests to check that various auxiliary routines are working."""

    def test_strain_increment(self):
        """Test for accumulating strain via strain increment calculations."""
        _, get_velocity_gradient = _velocity.simple_shear_2d("X", "Z", 1)
        timestamps = np.linspace(0, 1, 10)  # Solve until D₀t=1 (tensorial strain).
        strains_inc = np.zeros_like(timestamps)
        L = get_velocity_gradient(np.nan, Ŋ([0e0, 0e0, 0e0]))
        for i, ε in enumerate(strains_inc[1:]):
            strains_inc[i + 1] = strains_inc[i] + _utils.strain_increment(
                timestamps[1] - timestamps[0],
                L,
            )
        # For constant timesteps, check strains == positive_timestamps * strain_rate.
        nt.assert_allclose(strains_inc, timestamps, atol=6e-16, rtol=0)

        # Same thing, but for strain rate similar to experiments.
        _, get_velocity_gradient = _velocity.simple_shear_2d("Y", "X", 1e-5)
        timestamps = np.linspace(0, 1e6, 10)  # Solve until D₀t=10 (tensorial strain).
        strains_inc = np.zeros_like(timestamps)
        L = get_velocity_gradient(np.nan, Ŋ([0e0, 0e0, 0e0]))
        for i, ε in enumerate(strains_inc[1:]):
            strains_inc[i + 1] = strains_inc[i] + _utils.strain_increment(
                timestamps[1] - timestamps[0],
                L,
            )
        nt.assert_allclose(strains_inc, timestamps * 1e-5, atol=5e-15, rtol=0)

        # Again, but this time the particle will move (using get_pathline).
        # We use a 400km x 400km box and a strain rate of 1e-15 s⁻¹.
        get_velocity, get_velocity_gradient = _velocity.simple_shear_2d("X", "Z", 1e-15)
        timestamps, get_position = _paths.get_pathline(
            Ŋ([1e5, 0e0, 1e5]),
            get_velocity,
            get_velocity_gradient,
            Ŋ([-2e5, 0e0, -2e5]),
            Ŋ([2e5, 0e0, 2e5]),
            2,
            regular_steps=10,
        )
        positions = [get_position(t) for t in timestamps]
        velocity_gradients = [get_velocity_gradient(np.nan, Ŋ(x)) for x in positions]

        # Check that polycrystal is experiencing steady velocity gradient.
        nt.assert_array_equal(
            velocity_gradients, np.full_like(velocity_gradients, velocity_gradients[0])
        )
        # Check that positions are changing as expected.
        xdiff = np.diff(Ŋ([x[0] for x in positions]))
        zdiff = np.diff(Ŋ([x[2] for x in positions]))
        assert xdiff[0] > 0
        assert zdiff[0] == 0
        nt.assert_allclose(xdiff, np.full_like(xdiff, xdiff[0]), rtol=0, atol=1e-10)
        nt.assert_allclose(zdiff, np.full_like(zdiff, zdiff[0]), rtol=0, atol=1e-10)
        strains_inc = np.zeros_like(timestamps)
        for t, time in enumerate(timestamps[:-1], start=1):
            strains_inc[t] = strains_inc[t - 1] + (
                _utils.strain_increment(timestamps[t] - time, velocity_gradients[t])
            )
        # fig, ax, _, _ = _vis.pathline_box2d(
        #     None,
        #     get_velocity,
        #     "xz",
        #     strains_inc,
        #     positions,
        #     ".",
        #     Ŋ([-2e5, -2e5]),
        #     Ŋ([2e5, 2e5]),
        #     [20, 20],
        # )
        # fig.savefig("/tmp/fig.png")
        nt.assert_allclose(
            strains_inc,
            (timestamps - timestamps[0]) * 1e-15,
            atol=5e-15,
            rtol=0,
        )


class TestOlivineA:
    """Tests for stationary A-type olivine polycrystals in 2D simple shear."""

    class_id = "olivineA"

    @classmethod
    def run(
        cls,
        params,
        timestamps,
        strain_rate,
        get_velocity_gradient,
        shear_direction,
        seed=None,
        return_fse=None,
        get_position=lambda t: np.zeros(3),  # Stationary particles by default.
    ):
        """Reusable logic for 2D olivine (A-type) simple shear tests.

        Returns a tuple with the mineral and the FSE angle (or `None` if `return_fse` is
        `None`).

        """
        mineral = _minerals.Mineral(
            phase=_core.MineralPhase.olivine,
            fabric=_core.MineralFabric.olivine_A,
            regime=_core.DeformationRegime.dislocation,
            n_grains=params["number_of_grains"],
            seed=seed,
        )
        deformation_gradient = np.eye(3)  # Undeformed initial state.
        θ_fse = np.empty_like(timestamps)
        θ_fse[0] = 45

        for t, time in enumerate(timestamps[:-1], start=1):
            # Set up logging message depending on dynamic parameter and seeds.
            msg_start = (
                f"N = {params['number_of_grains']}; "
                + f"λ∗ = {params['nucleation_efficiency']}; "
                + f"X = {params['gbs_threshold']}; "
                + f"M∗ = {params['gbm_mobility']}; "
            )
            if seed is not None:
                msg_start += f"# {seed}; "

            _log.info(msg_start + "step %s/%s (t = %s)", t, len(timestamps) - 1, time)

            deformation_gradient = mineral.update_orientations(
                params,
                deformation_gradient,
                get_velocity_gradient,
                pathline=(time, timestamps[t], get_position),
            )
            _log.debug(
                "› velocity gradient = %s",
                get_velocity_gradient(np.nan, np.full(3, np.nan)).flatten(),
            )
            _log.debug("› strain D₀t = %.2f", strain_rate * timestamps[t])
            _log.debug(
                "› grain fractions: median = %s, max = %s, min = %s",
                np.median(mineral.fractions[-1]),
                np.max(mineral.fractions[-1]),
                np.min(mineral.fractions[-1]),
            )
            if return_fse:
                _, fse_v = _diagnostics.finite_strain(deformation_gradient)
                θ_fse[t] = _diagnostics.smallest_angle(fse_v, shear_direction)
            else:
                θ_fse = None

        return mineral, θ_fse

    @classmethod
    def interp_GBM_Kaminski2001(cls, strains):
        """Interpolate Kaminski & Ribe, 2001 data to get target angles at `strains`."""
        _log.info("interpolating target CPO angles...")
        data = _io.read_scsv(_io.data("thirdparty") / "Kaminski2001_GBMshear.scsv")
        cs_M0 = PchipInterpolator(
            _utils.remove_nans(data.equivalent_strain_M0) / 200,
            _utils.remove_nans(data.angle_M0),
        )
        cs_M50 = PchipInterpolator(
            _utils.remove_nans(data.equivalent_strain_M50) / 200,
            _utils.remove_nans(data.angle_M50),
        )
        cs_M200 = PchipInterpolator(
            _utils.remove_nans(data.equivalent_strain_M200) / 200,
            _utils.remove_nans(data.angle_M200),
        )
        return [cs_M0(strains), cs_M50(strains), cs_M200(strains)]

    @classmethod
    def interp_GBM_FortranDRex(cls, strains):
        """Interpolate angles produced using 'tools/drex_forward_simpleshear.f90'."""
        _log.info("interpolating target CPO  angles...")
        data = _io.read_scsv(_io.data("drexF90") / "olA_D1E4_dt50_X0_L5.scsv")
        data_strains = np.linspace(0, 1, 200)
        cs_M0 = PchipInterpolator(data_strains, _utils.remove_nans(data.M0_angle))
        cs_M50 = PchipInterpolator(data_strains, _utils.remove_nans(data.M50_angle))
        cs_M200 = PchipInterpolator(data_strains, _utils.remove_nans(data.M200_angle))
        return [cs_M0(strains), cs_M50(strains), cs_M200(strains)]

    @classmethod
    def interp_GBS_FortranDRex(cls, strains):
        """Interpolate angles produced using 'tools/drex_forward_simpleshear.f90'."""
        _log.info("interpolating target CPO  angles...")
        data = _io.read_scsv(_io.data("thirdparty") / "a_axis_GBS_fortran.scsv")
        data_strains = np.linspace(0, 1, 200)
        cs_X0 = PchipInterpolator(data_strains, _utils.remove_nans(data.a_mean_X0))
        cs_X20 = PchipInterpolator(data_strains, _utils.remove_nans(data.a_mean_X20))
        cs_X40 = PchipInterpolator(data_strains, _utils.remove_nans(data.a_mean_X40))
        return [cs_X0(strains), cs_X20(strains), cs_X40(strains)]

    @classmethod
    def interp_GBS_long_FortranDRex(cls, strains):
        """Interpolate angles produced using 'tools/drex_forward_simpleshear.f90'."""
        _log.info("interpolating target CPO  angles...")
        data = _io.read_scsv(_io.data("thirdparty") / "a_axis_GBS_long_fortran.scsv")
        data_strains = np.linspace(0, 2.5, 500)
        cs_X0 = PchipInterpolator(data_strains, _utils.remove_nans(data.a_mean_X0))
        cs_X20 = PchipInterpolator(data_strains, _utils.remove_nans(data.a_mean_X20))
        cs_X40 = PchipInterpolator(data_strains, _utils.remove_nans(data.a_mean_X40))
        return [cs_X0(strains), cs_X20(strains), cs_X40(strains)]

    @classmethod
    def interp_GBS_Kaminski2004(cls, strains):
        """Interpolate Kaminski & Ribe, 2001 data to get target angles at `strains`."""
        _log.info("interpolating target CPO angles...")
        data = _io.read_scsv(_io.data("thirdparty") / "Kaminski2004_GBSshear.scsv")
        cs_X0 = PchipInterpolator(
            _utils.remove_nans(data.dimensionless_time_X0),
            45 + _utils.remove_nans(data.angle_X0),
        )
        cs_X0d2 = PchipInterpolator(
            _utils.remove_nans(data.dimensionless_time_X0d2),
            45 + _utils.remove_nans(data.angle_X0d2),
        )
        cs_X0d4 = PchipInterpolator(
            _utils.remove_nans(data.dimensionless_time_X0d4),
            45 + _utils.remove_nans(data.angle_X0d4),
        )
        return [cs_X0(strains), cs_X0d2(strains), cs_X0d4(strains)]

    @pytest.mark.skipif(sys.platform == "win32", reason="Unable to allocate memory")
    def test_zero_recrystallisation(self, seed):
        """Check that M*=0 is a reliable switch to turn off recrystallisation."""
        params = _io.DEFAULT_PARAMS
        params["gbm_mobility"] = 0
        strain_rate = 1
        timestamps = np.linspace(0, 1, 25)  # Solve until D₀t=1 (tensorial strain).
        shear_direction = Ŋ([0, 1, 0], dtype=np.float64)
        _, get_velocity_gradient = _velocity.simple_shear_2d("Y", "X", strain_rate)
        mineral, _ = self.run(
            params,
            timestamps,
            strain_rate,
            get_velocity_gradient,
            shear_direction,
            seed=seed,
        )
        for fractions in mineral.fractions[1:]:
            nt.assert_allclose(fractions, mineral.fractions[0], atol=1e-15, rtol=0)

    @pytest.mark.skipif(sys.platform == "win32", reason="Unable to allocate memory")
    @pytest.mark.parametrize("gbm_mobility", [50, 100, 150])
    def test_grainsize_median(self, seed, gbm_mobility):
        """Check that M*={50,100,150}, λ*=5 causes decreasing grain size median."""
        params = _io.DEFAULT_PARAMS
        params["gbm_mobility"] = gbm_mobility
        params["nucleation_efficiency"] = 5
        strain_rate = 1
        timestamps = np.linspace(0, 1, 25)  # Solve until D₀t=1 (tensorial strain).
        n_timestamps = len(timestamps)
        shear_direction = Ŋ([0, 1, 0], dtype=np.float64)
        _, get_velocity_gradient = _velocity.simple_shear_2d("Y", "X", strain_rate)
        mineral, _ = self.run(
            params,
            timestamps,
            strain_rate,
            get_velocity_gradient,
            shear_direction,
            seed=seed,
        )
        medians = np.empty(n_timestamps)
        for i, fractions in enumerate(mineral.fractions):
            medians[i] = np.median(fractions)

        # The first diff is positive (~1e-6) before nucleation sets in.
        nt.assert_array_less(np.diff(medians)[1:], np.full(n_timestamps - 2, 0))

    @pytest.mark.slow
    @pytest.mark.parametrize("gbs_threshold", [0, 0.2, 0.4])
    @pytest.mark.parametrize("nucleation_efficiency", [3, 5, 10])
    def test_dvdx_ensemble(
        self, outdir, seeds_nearX45, ncpus, gbs_threshold, nucleation_efficiency
    ):
        r"""Test a-axis alignment to shear in Y direction (init. SCCS near 45° from X).

        Velocity gradient:
        $$\bm{L} = \begin{bmatrix} 0 & 0 & 0 \cr 2 & 0 & 0 \cr 0 & 0 & 0 \end{bmatrix}$$

        """
        strain_rate = 1
        timestamps = np.linspace(0, 1, 201)  # Solve until D₀t=1 ('shear' γ=2).
        n_timestamps = len(timestamps)
        # Use `seeds` instead of `seeds_nearX45` if you have even more RAM and CPU time.
        _seeds = seeds_nearX45
        n_seeds = len(_seeds)

        shear_direction = Ŋ([0, 1, 0], dtype=np.float64)
        _, get_velocity_gradient = _velocity.simple_shear_2d("Y", "X", strain_rate)

        gbm_mobilities = [0, 50, 125, 200]
        markers = ("x", "*", "d", "s")

        _id = f"X{_io.stringify(gbs_threshold)}_L{_io.stringify(nucleation_efficiency)}"
        # Output setup with optional logging and data series labels.
        θ_fse = np.empty_like(timestamps)
        angles = np.empty((len(gbm_mobilities), n_seeds, n_timestamps))
        optional_logging = cl.nullcontext()
        if outdir is not None:
            out_basepath = f"{outdir}/{SUBDIR}/{self.class_id}_dvdx_ensemble_{_id}"
            optional_logging = _log.logfile_enable(f"{out_basepath}.log")
            labels = []

        with optional_logging:
            clock_start = process_time()
            for m, gbm_mobility in enumerate(gbm_mobilities):
                if m == 0:
                    return_fse = True
                else:
                    return_fse = False

                params = {
                    "olivine_fraction": 1.0,
                    "enstatite_fraction": 0.0,
                    "stress_exponent": 1.5,
                    "deformation_exponent": 3.5,
                    "gbm_mobility": gbm_mobility,
                    "gbs_threshold": gbs_threshold,
                    "nucleation_efficiency": nucleation_efficiency,
                    "number_of_grains": 5000,
                    "initial_olivine_fabric": "A",
                }

                _run = ft.partial(
                    self.run,
                    params,
                    timestamps,
                    strain_rate,
                    get_velocity_gradient,
                    shear_direction,
                    return_fse=return_fse,
                )
                with Pool(processes=ncpus) as pool:
                    for s, out in enumerate(pool.imap_unordered(_run, _seeds)):
                        mineral, fse_angles = out
                        angles[m, s, :] = [
                            _diagnostics.smallest_angle(v, shear_direction)
                            for v in _diagnostics.elasticity_components(
                                _minerals.voigt_averages([mineral], params)
                            )["hexagonal_axis"]
                        ]
                        # Save the whole mineral for the first seed only.
                        if outdir is not None and s == 0:
                            postfix = (
                                f"M{_io.stringify(gbm_mobility)}"
                                + f"_X{_io.stringify(gbs_threshold)}"
                                + f"_L{_io.stringify(nucleation_efficiency)}"
                            )
                            mineral.save(f"{out_basepath}.npz", postfix=postfix)
                        if return_fse:
                            θ_fse += fse_angles

                if return_fse:
                    θ_fse /= n_seeds

                if outdir is not None:
                    labels.append(f"$M^∗$ = {gbm_mobility}")

            _log.info("elapsed CPU time: %s", np.abs(process_time() - clock_start))

        # Take ensemble means and optionally plot figure.
        strains = timestamps * strain_rate
        _log.info("postprocessing results for %s", _id)
        result_angles = angles.mean(axis=1)
        result_angles_err = angles.std(axis=1)

        if outdir is not None:
            schema = {
                "delimiter": ",",
                "missing": "-",
                "fields": [
                    {
                        "name": "strain",
                        "type": "integer",
                        "unit": "percent",
                        "fill": 999999,
                    }
                ],
            }
            np.savez(
                f"{out_basepath}.npz",
                angles=result_angles,
                angles_err=result_angles_err,
            )
            _io.save_scsv(
                f"{out_basepath}_strains.scsv",
                schema,
                [[int(D * 200) for D in strains]],  # Shear strain % is 200 * D₀.
            )
            fig, ax, colors = _vis.alignment(
                None,
                strains,
                result_angles,
                markers,
                labels,
                err=result_angles_err,
                θ_max=60,
                θ_fse=θ_fse,
            )
            fig.savefig(_io.resolve_path(f"{out_basepath}.pdf"))

    @pytest.mark.slow
    def test_dvdx_GBM(self, outdir, seeds_nearX45, ncpus):
        r"""Test a-axis alignment to shear in Y direction (init. SCCS near 45° from X).

        Velocity gradient:
        $$
        \bm{L} = 10^{-4} ×
            \begin{bmatrix} 0 & 0 & 0 \cr 2 & 0 & 0 \cr 0 & 0 & 0 \end{bmatrix}
        $$

        Results are compared to the Fortran 90 output.

        """
        shear_direction = Ŋ([0, 1, 0], dtype=np.float64)
        strain_rate = 1e-4
        _, get_velocity_gradient = _velocity.simple_shear_2d("Y", "X", strain_rate)
        timestamps = np.linspace(0, 1e4, 51)  # Solve until D₀t=1 ('shear' γ=2).
        i_strain_40p = 10  # Index of 40% strain, lower strains are not relevant here.
        i_strain_100p = 25  # Index of 100% strain, when M*=0 matches FSE.
        params = _io.DEFAULT_PARAMS
        params["gbs_threshold"] = 0  # No GBS, to match the Fortran parameters.
        gbm_mobilities = (0, 10, 50, 125, 200)  # Must be in ascending order.
        markers = ("x", ".", "*", "d", "s")
        # Use `seeds` instead of `seeds_nearX45` if you have even more RAM and CPU time.
        _seeds = seeds_nearX45
        n_seeds = len(_seeds)
        angles = np.empty((len(gbm_mobilities), n_seeds, len(timestamps)))
        θ_fse = np.zeros_like(timestamps)
        strains = timestamps * strain_rate
        M0_drexF90, M50_drexF90, M200_drexF90 = self.interp_GBM_FortranDRex(strains)

        optional_logging = cl.nullcontext()
        if outdir is not None:
            out_basepath = f"{outdir}/{SUBDIR}/{self.class_id}_mobility"
            optional_logging = _log.logfile_enable(f"{out_basepath}.log")
            labels = []

        with optional_logging:
            clock_start = process_time()
            for m, gbm_mobility in enumerate(gbm_mobilities):
                if m == 0:
                    return_fse = True
                else:
                    return_fse = False
                params["gbm_mobility"] = gbm_mobility

                _run = ft.partial(
                    self.run,
                    params,
                    timestamps,
                    strain_rate,
                    get_velocity_gradient,
                    shear_direction,
                    return_fse=True,
                )
                with Pool(processes=ncpus) as pool:
                    for s, out in enumerate(pool.imap_unordered(_run, _seeds)):
                        mineral, fse_angles = out
                        angles[m, s, :] = [
                            _diagnostics.smallest_angle(v, shear_direction)
                            for v in _diagnostics.elasticity_components(
                                _minerals.voigt_averages([mineral], params)
                            )["hexagonal_axis"]
                        ]
                        # Save the whole mineral for the first seed only.
                        if outdir is not None and s == 0:
                            mineral.save(
                                f"{out_basepath}.npz",
                                postfix=f"M{_io.stringify(gbm_mobility)}",
                            )
                        if return_fse:
                            θ_fse += fse_angles

                if return_fse:
                    θ_fse /= n_seeds

                if outdir is not None:
                    labels.append(f"$M^∗$ = {params['gbm_mobility']}")

            _log.info("elapsed CPU time: %s", np.abs(process_time() - clock_start))

            # Take ensemble means and optionally plot figure.
            _log.info("postprocessing results...")
            result_angles = angles.mean(axis=1)
            result_angles_err = angles.std(axis=1)

            if outdir is not None:
                schema = {
                    "delimiter": ",",
                    "missing": "-",
                    "fields": [
                        {
                            "name": "strain",
                            "type": "integer",
                            "unit": "percent",
                            "fill": 999999,
                        }
                    ],
                }
                _io.save_scsv(
                    f"{out_basepath}_strains.scsv",
                    schema,
                    [[int(D * 200) for D in strains]],  # Shear strain % is 200 * D₀.
                )
                np.savez(
                    f"{out_basepath}_angles.npz",
                    angles=result_angles,
                    err=result_angles_err,
                )
                fig, ax, colors = _vis.alignment(
                    None,
                    strains,
                    result_angles,
                    markers,
                    labels,
                    err=result_angles_err,
                    θ_max=60,
                    θ_fse=θ_fse,
                )
                ax.plot(strains, M0_drexF90, c=colors[0])
                ax.plot(strains, M50_drexF90, c=colors[2])
                ax.plot(strains, M200_drexF90, c=colors[4])
                _vis.show_Skemer2016_ShearStrainAngles(
                    ax,
                    ["Z&K 1200 C", "Z&K 1300 C"],
                    ["v", "^"],
                    ["k", "k"],
                    ["none", None],
                    [
                        "Zhang & Karato, 1995\n(1473 K)",
                        "Zhang & Karato, 1995\n(1573 K)",
                    ],
                    _core.MineralFabric.olivine_A,
                )
                # There is a lot of stuff on this legend, so put it outside the axes.
                # These values might need to be tweaked depending on the font size, etc.
                _legend = _utils.redraw_legend(ax, fig=fig, bbox_to_anchor=(1.66, 0.99))
                fig.savefig(
                    _io.resolve_path(f"{out_basepath}.pdf"),
                    bbox_extra_artists=(_legend,),
                    bbox_inches="tight",
                )

            # Check that GBM speeds up the alignment between 40% and 100% strain.
            _log.info("checking grain orientations...")
            for i, θ in enumerate(result_angles[:-1], start=1):
                nt.assert_array_less(
                    result_angles[i][i_strain_40p:i_strain_100p],
                    θ[i_strain_40p:i_strain_100p],
                )

            # Check that M*=0 matches FSE (±1°) past 100% strain.
            nt.assert_allclose(
                result_angles[0][i_strain_100p:],
                θ_fse[i_strain_100p:],
                atol=1,
                rtol=0,
            )

            # Check that results match Fortran output past 40% strain.
            nt.assert_allclose(
                result_angles[0][i_strain_40p:],
                M0_drexF90[i_strain_40p:],
                atol=0,
                rtol=0.1,  # At 40% strain the match is worse than at higher strain.
            )
            nt.assert_allclose(
                result_angles[2][i_strain_40p:],
                M50_drexF90[i_strain_40p:],
                atol=1,
                rtol=0,
            )
            nt.assert_allclose(
                result_angles[4][i_strain_40p:],
                M200_drexF90[i_strain_40p:],
                atol=1.5,
                rtol=0,
            )

    @pytest.mark.slow
    def test_GBM_calibration(self, outdir, seeds, ncpus):
        r"""Compare results for various values of $$M^∗$$ to A-type olivine data.

        Velocity gradient:
        $$
        \bm{L} = 10^{-4} ×
            \begin{bmatrix} 0 & 0 & 0 \cr 2 & 0 & 0 \cr 0 & 0 & 0 \end{bmatrix}
        $$

        Unlike `test_dvdx_GBM`,
        grain boudary sliding is enabled here (see `_io.DEFAULT_PARAMS`).
        Data are provided by [Skemer & Hansen, 2016](http://dx.doi.org/10.1016/j.tecto.2015.12.003).

        """
        shear_direction = Ŋ([0, 1, 0], dtype=np.float64)
        strain_rate = 1
        _, get_velocity_gradient = _velocity.simple_shear_2d("Y", "X", strain_rate)
        timestamps = np.linspace(0, 3.2, 65)  # Solve until D₀t=3.2 ('shear' γ=6.4).
        params = _io.DEFAULT_PARAMS
        params["number_of_grains"] = 5000
        gbm_mobilities = (0, 10, 50, 125)  # Must be in ascending order.
        markers = ("x", "*", "1", ".")
        # Uses 100 seeds by default; use all 1000 if you have more RAM and CPU time.
        _seeds = seeds[:100]
        n_seeds = len(_seeds)
        angles = np.empty((len(gbm_mobilities), n_seeds, len(timestamps)))
        θ_fse = np.zeros_like(timestamps)
        strains = timestamps * strain_rate

        optional_logging = cl.nullcontext()
        if outdir is not None:
            out_basepath = f"{outdir}/{SUBDIR}/{self.class_id}_calibration"
            optional_logging = _log.logfile_enable(f"{out_basepath}.log")
            labels = []

        with optional_logging:
            clock_start = process_time()
            for m, gbm_mobility in enumerate(gbm_mobilities):
                return_fse = True if m == 0 else False
                params["gbm_mobility"] = gbm_mobility
                _run = ft.partial(
                    self.run,
                    params,
                    timestamps,
                    strain_rate,
                    get_velocity_gradient,
                    shear_direction,
                    return_fse=return_fse,
                )
                with Pool(processes=ncpus) as pool:
                    for s, out in enumerate(pool.imap_unordered(_run, _seeds)):
                        mineral, fse_angles = out
                        angles[m, s, :] = [
                            _diagnostics.smallest_angle(v, shear_direction)
                            for v in _diagnostics.elasticity_components(
                                _minerals.voigt_averages([mineral], params)
                            )["hexagonal_axis"]
                        ]
                        # Save the whole mineral for the first seed only.
                        if outdir is not None and s == 0:
                            mineral.save(
                                f"{out_basepath}.npz",
                                postfix=f"M{_io.stringify(gbm_mobility)}",
                            )
                        if return_fse:
                            θ_fse += fse_angles

                if return_fse:
                    θ_fse /= n_seeds

                if outdir is not None:
                    labels.append(f"$M^∗$ = {params['gbm_mobility']}")

            _log.info("elapsed CPU time: %s", np.abs(process_time() - clock_start))

            # Take ensemble means and optionally plot figure.
            _log.info("postprocessing results...")
            result_angles = angles.mean(axis=1)
            result_angles_err = angles.std(axis=1)

            if outdir is not None:
                schema = {
                    "delimiter": ",",
                    "missing": "-",
                    "fields": [
                        {
                            "name": "strain",
                            "type": "integer",
                            "unit": "percent",
                            "fill": 999999,
                        }
                    ],
                }
                _io.save_scsv(
                    f"{out_basepath}_strains.scsv",
                    schema,
                    [[int(D * 200) for D in strains]],  # Shear strain % is 200 * D₀.
                )
                np.savez(
                    _io.resolve_path(f"{out_basepath}_ensemble_means.npz"),
                    angles=result_angles,
                    err=result_angles_err,
                )
                fig = _vis.figure(
                    figsize=(_vis.DEFAULT_FIG_WIDTH * 3, _vis.DEFAULT_FIG_HEIGHT)
                )
                fig, ax, colors = _vis.alignment(
                    fig.add_subplot(),
                    strains,
                    result_angles,
                    markers,
                    labels,
                    err=result_angles_err,
                    θ_max=80,
                    θ_fse=θ_fse,
                )
                (
                    _,
                    _,
                    _,
                    data_Skemer2016,
                    indices,
                ) = _vis.show_Skemer2016_ShearStrainAngles(
                    ax,
                    [
                        "Z&K 1200 C",
                        "Z&K 1300 C",
                        "Skemer 2011",
                        "Hansen 2014",
                        "Warren 2008",
                        "Webber 2010",
                        "H&W 2015",
                    ],
                    ["v", "^", "o", "s", "v", "o", "s"],
                    ["k", "k", "k", "k", "k", "k", "k"],
                    ["none", "none", "none", "none", None, None, None],
                    [
                        "Zhang & Karato, 1995 (1473 K)",
                        "Zhang & Karato, 1995 (1573 K)",
                        "Skemer et al., 2011 (1500 K)",
                        "Hansen et al., 2014 (1473 K)",
                        "Warren et al., 2008",
                        "Webber et al., 2010",
                        "Hansen & Warren, 2015",
                    ],
                    fabric=_core.MineralFabric.olivine_A,
                )
                _legend = _utils.redraw_legend(ax, loc="upper right", ncols=3)
                fig.savefig(
                    _io.resolve_path(f"{out_basepath}.pdf"),
                    bbox_extra_artists=(_legend,),
                    bbox_inches="tight",
                )
                r2vals = []
                for angles in result_angles:
                    _angles = PchipInterpolator(strains, angles)
                    r2 = np.sum(
                        [
                            (a - b) ** 2
                            for a, b in zip(
                                _angles(
                                    np.take(data_Skemer2016.shear_strain, indices) / 200
                                ),
                                np.take(data_Skemer2016.angle, indices),
                            )
                        ]
                    )
                    r2vals.append(r2)
                _log.info(
                    "Sums of squared residuals (r-values) for each M∗: %s", r2vals
                )

    @pytest.mark.big
    def test_dudz_pathline(self, outdir, seed):
        """Test alignment of olivine a-axis for a polycrystal advected on a pathline."""
        test_id = "dudz_pathline"
        optional_logging = cl.nullcontext()
        if outdir is not None:
            out_basepath = f"{outdir}/{SUBDIR}/{self.class_id}_{test_id}"
            optional_logging = _log.logfile_enable(f"{out_basepath}.log")

        with optional_logging:
            shear_direction = Ŋ([1, 0, 0], dtype=np.float64)
            strain_rate = 1e-15  # Moderate, realistic shear in the upper mantle.
            get_velocity, get_velocity_gradient = _velocity.simple_shear_2d(
                "X", "Z", strain_rate
            )
            n_timesteps = 10
            timestamps, get_position = _paths.get_pathline(
                Ŋ([1e5, 0e0, 1e5]),
                get_velocity,
                get_velocity_gradient,
                Ŋ([-2e5, 0e0, -2e5]),
                Ŋ([2e5, 0e0, 2e5]),
                2,
                regular_steps=n_timesteps,
            )
            positions = [get_position(t) for t in timestamps]
            velocity_gradients = [
                get_velocity_gradient(np.nan, Ŋ(x)) for x in positions
            ]

            params = _io.DEFAULT_PARAMS
            params["gbm_mobility"] = 10
            params["number_of_grains"] = 5000
            olA = _minerals.Mineral(n_grains=params["number_of_grains"], seed=seed)
            deformation_gradient = np.eye(3)
            strains = np.zeros_like(timestamps)
            for t, time in enumerate(timestamps[:-1], start=1):
                strains[t] = strains[t - 1] + (
                    _utils.strain_increment(timestamps[t] - time, velocity_gradients[t])
                )
                _log.info("step %d/%d (ε = %.2f)", t, len(timestamps) - 1, strains[t])
                deformation_gradient = olA.update_orientations(
                    params,
                    deformation_gradient,
                    get_velocity_gradient,
                    pathline=(time, timestamps[t], get_position),
                )

            if outdir is not None:
                olA.save(f"{out_basepath}.npz")

            orient_resampled, fractions_resampled = _stats.resample_orientations(
                olA.orientations, olA.fractions, seed=seed
            )
            # About 36GB, 26 min needed with float64. GitHub macos runner has 14GB.
            misorient_indices = _diagnostics.misorientation_indices(
                orient_resampled,
                _geo.LatticeSystem.orthorhombic,
                ncpus=3,
            )
            cpo_vectors = np.zeros((n_timesteps + 1, 3))
            cpo_angles = np.zeros(n_timesteps + 1)
            for i, matrices in enumerate(orient_resampled):
                cpo_vectors[i] = _diagnostics.bingham_average(
                    matrices,
                    axis=_minerals.OLIVINE_PRIMARY_AXIS[olA.fabric],
                )
                cpo_angles[i] = _diagnostics.smallest_angle(
                    cpo_vectors[i], shear_direction
                )

            # Check for mostly decreasing CPO angles (exclude initial condition).
            _log.debug("cpo angles: %s", cpo_angles)
            nt.assert_array_less(np.diff(cpo_angles[1:]), np.ones(n_timesteps - 1))
            # Check for increasing CPO strength (M-index).
            _log.debug("cpo strengths: %s", misorient_indices)
            nt.assert_array_less(
                np.full(n_timesteps, -0.01), np.diff(misorient_indices)
            )
            # Check that last angle is <5° (M*=125) or <10° (M*=10).
            assert cpo_angles[-1] < 10

        if outdir is not None:
            fig, ax, _, _ = _vis.pathline_box2d(
                None,
                get_velocity,
                "xz",
                strains,
                positions,
                ".",
                Ŋ([-2e5, -2e5]),
                Ŋ([2e5, 2e5]),
                [20, 20],
                cpo_vectors=cpo_vectors,
                cpo_strengths=misorient_indices,
            )
            fig.savefig(f"{out_basepath}.pdf")



from time import perf_counter

import numba as nb
import numpy as np

from pydrex import core as _core
from pydrex import diagnostics as _diagnostics
from pydrex import io as _io
from pydrex import minerals as _minerals
from pydrex import velocity as _velocity

shear_direction = np.array([0.0, 1.0, 0.0])
strain_rate = 1e-4
_, get_velocity_gradient = _velocity.simple_shear_2d("Y", "X", strain_rate)
timestamps = np.linspace(0, 1e4, 201)
params = _io.DEFAULT_PARAMS
params["number_of_grains"] = 15**3
params["gbs_threshold"] = 0
seed = 245623452

mineral = _minerals.Mineral(
    phase=_core.MineralPhase.olivine,
    fabric=_core.MineralFabric.olivine_A,
    regime=_core.DeformationRegime.dislocation,
    n_grains=params["number_of_grains"],
    seed=seed,
)
deformation_gradient = np.eye(3)
θ_fse = np.empty_like(timestamps)
θ_fse[0] = 45


@nb.njit
def get_position(t):
    return np.zeros(3)


# Warm up run to make numba compile things.
mineral.update_orientations(
    params,
    deformation_gradient,
    get_velocity_gradient,
    pathline=(timestamps[0], timestamps[1], get_position),
)

_begin = perf_counter()
for t, time in enumerate(timestamps[:-1], start=1):
    deformation_gradient = mineral.update_orientations(
        params,
        deformation_gradient,
        get_velocity_gradient,
        pathline=(time, timestamps[t], get_position),
    )
    _, fse_v = _diagnostics.finite_strain(deformation_gradient)
    θ_fse[t] = _diagnostics.smallest_angle(fse_v, shear_direction)

# Cij_and_friends = _diagnostics.elasticity_components(
#     _minerals.voigt_averages([mineral], params)
# )
# angles = [
#     _diagnostics.smallest_angle(v, shear_direction)
#     for v in Cij_and_friends["hexagonal_axis"]
# ]
# percent_anisotropy = Cij_and_friends["percent_anisotropy"]

print(f"Elapsed time: {perf_counter() - _begin}")



import configparser
import pathlib

import numba as nb
import numpy as np

from pydrex.utils import add_dim, remove_dim
from pydrex.velocity import simple_shear_2d

config = configparser.ConfigParser()
config.read(pathlib.PurePath(__file__).with_suffix(".ini"))
init = config["initial conditions"]


def init_positions():
    """Return initial particle positions."""
    init_horiz = np.array([float(x) for x in init["INIT_HORIZ"].split()])
    init_vert = float(init["INIT_VERT"])
    return np.array([[x, init_vert] for x in init_horiz])


def cb_interp_position(position_prev, position, t, dt):
    """Return position interpolator for CPO integration from `position_prev` to `position`."""

    @nb.njit(fastmath=True)
    def _interp_position(int_time):
        return position_prev + (int_time - t + dt) / dt * (position - position_prev)

    return _interp_position


def cb_interp_velocity_grad(velocity_grad_prev, velocity_grad, t, dt):
    """Return ∇v interpolator for CPO integration between `velocity_grad_prev` and `velocity_grad`."""

    @nb.njit(fastmath=True)
    def _interp_vel_grad(int_time, int_position):
        return velocity_grad_prev + (int_time - t + dt) / dt * (
            velocity_grad - velocity_grad_prev
        )

    return _interp_vel_grad


STRAIN_RATE = float(init["STRAIN_RATE"])
# Initial position of the particle.
INITIAL_POSITIONS = init_positions()
# Recrystallisation parameter configuration.
PARAMS = {
    "stress_exponent": float(init["STRESS_EXPONENT"]),
    "deformation_exponent": float(init["DEFORMATION_EXPONENT"]),
    "gbm_mobility": float(init["GBM_MOBILITY"]),
    "nucleation_efficiency": float(init["NUCLEATION_EFFICIENCY"]),
    "gbs_threshold": float(init["GBS_THRESHOLD"]),
    "minerals": ("olivine",),
    "olivine_fraction": 1,
}

_velocity, _velocity_grad = simple_shear_2d("X", "Z", STRAIN_RATE)


def velocity(t, X):
    return remove_dim(_velocity(t, add_dim(X, 1)), 1)


def velocity_gradient(t, X):
    return remove_dim(_velocity_grad(t, add_dim(X, 1)), 1)



"""> PyDRex: Functions for geometric coordinate conversions and projections."""

from enum import Enum, unique

import numba as nb
import numpy as np
from scipy import linalg as la
from scipy.spatial.transform import Rotation


@unique
class LatticeSystem(Enum):
    """Crystallographic lattice systems supported by postprocessing methods.

    The value of a member is `(a, b)` with `a` and `b` as given in the table below.
    The additional row lists the maximum misorientation angle between two crystallites
    for the given lattice system.

                triclinic  monoclinic  orthorhombic  rhombohedral tetragonal hexagonal
        ------------------------------------------------------------------------------
        a       1          2           2             3            4          6
        b       1          2           4             6            8          12
        θmax    180°       180°        120°          120°         90°        90°

    This is identically Table 1 in [Grimmer (1979)](https://doi.org/10.1016/0036-9748(79)90058-9).

    """

    triclinic = (1, 1)
    monoclinic = (2, 2)
    orthorhombic = (2, 4)
    rhombohedral = (3, 6)
    tetragonal = (4, 8)
    hexagonal = (6, 12)


def to_cartesian(ϕ, θ, r=1):
    """Convert spherical to cartesian coordinates in ℝ³.

    Spherical coordinate convention:
    - ϕ is the longitude (“azimuth”) ∈ [0, 2π)
    - θ is the colatitude (“inclination”) ∈ [0, π)

    By default, a radius of `r = 1` is used for the sphere.
    Returns a tuple containing arrays of x, y, and z values.

    """
    ϕ = np.atleast_1d(ϕ).astype(float)
    θ = np.atleast_1d(θ).astype(float)
    r = np.atleast_1d(r).astype(float)
    return (r * np.sin(θ) * np.cos(ϕ), r * np.sin(θ) * np.sin(ϕ), r * np.cos(θ))


def to_spherical(x, y, z):
    """Convert cartesian coordinates in ℝ³ to spherical coordinates.

    Spherical coordinate convention:
    - ϕ is the longitude (“azimuth”) ∈ [0, 2π)
    - θ is the colatitude (“inclination”) ∈ [0, π)

    Returns a tuple containing arrays of r, ϕ and θ values.

    """
    x = np.atleast_1d(x).astype(float)
    y = np.atleast_1d(y).astype(float)
    z = np.atleast_1d(z).astype(float)
    r = np.sqrt(x**2 + y**2 + z**2)
    return (r, np.arctan2(y, x), np.sign(y) * np.arccos(x / np.sqrt(x**2 + y**2)))


@nb.njit(fastmath=True)
def misorientation_angles(q1_array, q2_array):
    """Calculate minimum misorientation angles for collections of rotation quaternions.

    Calculate the smallest angular distance between any quaternions `q1_array[:, i]` and
    `q2_array[:, j]`, where i == j and the first dimensions of `q1_array` and `q2_array`
    are of equal length (the output will also be this long):

        q1_array.shape      q2_array.shape      len(output)
        ---------------------------------------------------
        NxAx4               NxBx4               N

    .. warning::
        This method must be able to allocate a floating point array of shape Nx(A*B)

    Uses ~25% less memory than the same operation with rotation matrices.

    See also:
    - <https://math.stackexchange.com/questions/90081/quaternion-distance>
    - [Huynh (2009)](https://link.springer.com/article/10.1007/s10851-009-0161-2)


    """
    if q1_array.shape[0] != q2_array.shape[0]:
        raise ValueError(
            "the first dimensions of q1_array and q2_array must be of equal length"
            + f" but {q1_array.shape[0]} != {q2_array.shape[0]}"
        )
    angles = np.empty((q1_array.shape[0], q1_array.shape[1] * q2_array.shape[1]))
    k = 0
    for i in range(q1_array.shape[1]):
        for j in range(q2_array.shape[1]):
            angles[:, k] = 2 * np.rad2deg(
                np.arccos(
                    np.abs(
                        np.clip(
                            np.sum(q1_array[:, i] * q2_array[:, j], axis=1),
                            -1.0,
                            1.0,
                        )
                    )
                )
            )
            k += 1
    return np.array([np.min(a) for a in angles])


def symmetry_operations(system: LatticeSystem):
    """Get sequence of symmetry operations for the given `LatticeSystem`.

    Returned transforms are either quaternions (for rotations of the lattice) or 4x4
    matrices which pre-multiply a quaternion to give a reflected variant (reflections
    are *improper* rotations and cannot be represented as quaternions or SciPy rotation
    matrices).

    """
    match system:
        case LatticeSystem.triclinic:  # Triclinic lattice, no additional symmetry.
            return [Rotation.identity().as_quat()]
        case LatticeSystem.monoclinic | LatticeSystem.orthorhombic:
            # Rotation by π around any of the Cartesian axes.
            rotations = [
                Rotation.from_rotvec(np.pi * np.asarray(vector)).as_quat()
                for vector in [[0, 0, 1], [0, 1, 0], [1, 0, 0]]
            ]
            # Reflections around any of the Cartesian planes.
            # Quaternion space matrix equivalents for the R^3 operations:
            #   np.diag(x) for x in ([1, 1, -1], [1, -1, 1], [-1, 1, 1])
            # https://stackoverflow.com/a/33999726/12519962
            # NOTE: Don't use scipy Rotation, it silently converts to proper rotations.
            # There is maybe <https://github.com/matthew-brett/transforms3d>,
            # but another dependency just for this might not be worth it.
            reflections = [
                np.diag(x) for x in ([1, -1, -1, 1], [1, -1, 1, -1], [1, 1, -1, -1])
            ]
            return [Rotation.identity().as_quat(), *rotations, *reflections]
        case LatticeSystem.rhombohedral:
            # Rotations by n * π/3, n ∈ {1, 2} around any of the Cartesian axes.
            rotations = [
                Rotation.from_rotvec(i * np.pi / 3 * np.asarray(vector)).as_quat()
                for vector in [[0, 0, 1], [0, 1, 0], [1, 0, 0]]
                for i in (1, 2)
            ]
            return [Rotation.identity().as_quat(), *rotations]
        case LatticeSystem.tetragonal:
            # Rotations by n * π/2, n ∈ {1, 2, 3} around any of the Cartesian axes.
            rotations = [
                Rotation.from_rotvec(i * np.pi / 2 * np.asarray(vector)).as_quat()
                for vector in [[0, 0, 1], [0, 1, 0], [1, 0, 0]]
                for i in (1, 2, 3)
            ]
            return [Rotation.identity().as_quat(), *rotations]
        case LatticeSystem.hexagonal:
            # Rotations by n * π/3, n ∈ {1, 2} around any of the Cartesian axes.
            rotations3 = [
                Rotation.from_rotvec(i * np.pi / 3 * np.asarray(vector)).as_quat()
                for vector in [[0, 0, 1], [0, 1, 0], [1, 0, 0]]
                for i in (1, 2)
            ]
            # Rotations by n * π/6, n ∈ {1, 3, 5} around any of the Cartesian axes.
            # The other three π/6 rotations are identical to the π/3 rotations.
            rotations6 = [
                Rotation.from_rotvec(i * np.pi / 6 * np.asarray(vector)).as_quat()
                for vector in [[0, 0, 1], [0, 1, 0], [1, 0, 0]]
                for i in (1, 3, 5)
            ]
            return [Rotation.identity().as_quat(), *rotations3, *rotations6]
        case _:
            raise ValueError(f"unsupported lattice system: {system}")


def poles(orientations, ref_axes="xz", hkl=[1, 0, 0]):
    """Extract 3D vectors of crystallographic directions from orientation matrices.

    Expects `orientations` to be an array with shape (N, 3, 3).
    The optional arguments `ref_axes` and `hkl` can be used to change
    the global reference axes and the crystallographic direction respectively.
    The reference axes should be given as a string of two letters,
    e.g. "xz" (default), which specify the second and third axes
    of the global right-handed reference frame. The third letter in the set "xyz"
    determines the first axis. The `ref_axes` will therefore become the
    horizontal and vertical axes of pole figures used to plot the directions.

    """
    _ref_axes = ref_axes.lower()
    upward_axes = (set("xyz") - set(_ref_axes)).pop()
    axes_map = {"x": 0, "y": 1, "z": 2}

    # Get directions in the right-handed frame.
    directions = np.tensordot(orientations.transpose([0, 2, 1]), hkl, axes=(2, 0))
    directions_norm = la.norm(directions, axis=1)
    directions /= directions_norm.reshape(-1, 1)

    # Rotate into the chosen reference frame.
    zvals = directions[:, axes_map[upward_axes]]
    yvals = directions[:, axes_map[_ref_axes[1]]]
    xvals = directions[:, axes_map[_ref_axes[0]]]
    return xvals, yvals, zvals


def lambert_equal_area(xvals, yvals, zvals):
    """Project axial data from a 3D sphere onto a 2D disk.

    Project points from a 3D sphere of radius 1, given in Cartesian coordinates,
    to points on a 2D disk using a Lambert equal area azimuthal projection.
    Returns arrays of the X and Y coordinates in the unit disk.

    This implementation first maps all points onto the same hemisphere,
    and then projects that hemisphere onto the disk.

    """
    xvals = np.atleast_1d(xvals).astype(float)
    yvals = np.atleast_1d(yvals).astype(float)
    zvals = np.atleast_1d(zvals).astype(float)
    # One reference for the equation is Mardia & Jupp 2009 (Directional Statistics),
    # where it appears as eq. 9.1.1 in spherical coordinates,
    #   [sinθcosφ, sinθsinφ, cosθ].

    # First we move all points into the upper hemisphere.
    # See e.g. page 186 of Snyder 1987 (Map Projections— A Working Manual).
    # This is done by taking θ' = π - θ where θ is the colatitude (inclination).
    # Equivalently, in Cartesian coords we just take abs of the z values.
    zvals = abs(zvals)
    # When x and y are both 0, we would have a zero-division.
    # These points are always projected onto [0, 0] (the centre of the disk).
    condition = np.logical_and(np.abs(xvals) < 1e-16, np.abs(yvals) < 1e-16)
    x_masked = np.ma.masked_where(condition, xvals)
    y_masked = np.ma.masked_where(condition, yvals)
    x_masked.fill_value = np.nan
    y_masked.fill_value = np.nan
    # The equations in Mardia & Jupp 2009 and Snyder 1987 both project the hemisphere
    # onto a disk of radius sqrt(2), so we drop the sqrt(2) factor that appears
    # after converting to Cartesian coords to get a radius of 1.
    prefactor = np.sqrt((1 - zvals) / (x_masked**2 + y_masked**2))
    prefactor.fill_value = 0.0
    return prefactor.filled() * xvals, prefactor.filled() * yvals


def shirley_concentric_squaredisk(xvals, yvals):
    """Project points from a square onto a disk using the concentric Shirley method.

    The concentric method of [Shirley & Chiu (1997)](https://doi.org/10.1080/10867651.1997.10487479)
    is optimised to preserve area. See also: <http://marc-b-reynolds.github.io/math/2017/01/08/SquareDisc.html>.

    This can be used to set up uniform grids on a disk, e.g.

    >>> a = [x / 5.0 for x in range(-5, 6)]
    >>> x = [[x] * len(a) for x in a]
    >>> y = [a for _ in a]
    >>> x_flat = [j for i in x for j in i]
    >>> y_flat = [j for i in y for j in i]
    >>> x_disk, y_disk = shirley_concentric_squaredisk(x_flat, y_flat)
    >>> r = x_disk**2 + y_disk**2
    >>> r.reshape((len(a), len(a)))
    array([[1.  , 1.  , 1.  , 1.  , 1.  , 1.  , 1.  , 1.  , 1.  , 1.  , 1.  ],
           [1.  , 0.64, 0.64, 0.64, 0.64, 0.64, 0.64, 0.64, 0.64, 0.64, 1.  ],
           [1.  , 0.64, 0.36, 0.36, 0.36, 0.36, 0.36, 0.36, 0.36, 0.64, 1.  ],
           [1.  , 0.64, 0.36, 0.16, 0.16, 0.16, 0.16, 0.16, 0.36, 0.64, 1.  ],
           [1.  , 0.64, 0.36, 0.16, 0.04, 0.04, 0.04, 0.16, 0.36, 0.64, 1.  ],
           [1.  , 0.64, 0.36, 0.16, 0.04, 0.  , 0.04, 0.16, 0.36, 0.64, 1.  ],
           [1.  , 0.64, 0.36, 0.16, 0.04, 0.04, 0.04, 0.16, 0.36, 0.64, 1.  ],
           [1.  , 0.64, 0.36, 0.16, 0.16, 0.16, 0.16, 0.16, 0.36, 0.64, 1.  ],
           [1.  , 0.64, 0.36, 0.36, 0.36, 0.36, 0.36, 0.36, 0.36, 0.64, 1.  ],
           [1.  , 0.64, 0.64, 0.64, 0.64, 0.64, 0.64, 0.64, 0.64, 0.64, 1.  ],
           [1.  , 1.  , 1.  , 1.  , 1.  , 1.  , 1.  , 1.  , 1.  , 1.  , 1.  ]])
    >>> from math import atan2
    >>> θ = [atan2(y, x) for y, x in zip(y_disk, x_disk)]
    >>> max(θ)
    3.141592653589793
    >>> min(θ)
    -2.9845130209101467

    """

    def _shirley_concentric_squaredisc_xgty(xvals, yvals):
        ratios = yvals / (xvals + 1e-12)
        return xvals * np.cos(np.pi / 4 * ratios), xvals * np.sin(np.pi / 4 * ratios)

    def _shirley_concentric_squaredisc_xlty(xvals, yvals):
        ratios = xvals / (yvals + 1e-12)
        return yvals * np.sin(np.pi / 4 * ratios), yvals * np.cos(np.pi / 4 * ratios)

    xvals = np.atleast_1d(xvals).astype(float)
    yvals = np.atleast_1d(yvals).astype(float)
    conditions = [
        np.repeat(np.atleast_2d(np.abs(xvals) >= np.abs(yvals)), 2, axis=0).transpose(),
        np.repeat(np.atleast_2d(np.abs(xvals) < np.abs(yvals)), 2, axis=0).transpose(),
    ]
    x_disk, y_disk = np.piecewise(
        np.column_stack([xvals, yvals]),
        conditions,
        [
            lambda xy: np.column_stack(
                _shirley_concentric_squaredisc_xgty(*xy.reshape(-1, 2).transpose())
            ).ravel(),
            lambda xy: np.column_stack(
                _shirley_concentric_squaredisc_xlty(*xy.reshape(-1, 2).transpose())
            ).ravel(),
        ],
    ).transpose()
    return x_disk, y_disk


def to_indices(horizontal, vertical):
    _geometry = (horizontal.upper(), vertical.upper())
    match _geometry:
        case ("X", "Y"):
            indices = (0, 1)
        case ("X", "Z"):
            indices = (0, 2)
        case ("Y", "X"):
            indices = (1, 0)
        case ("Y", "Z"):
            indices = (1, 2)
        case ("Z", "X"):
            indices = (2, 0)
        case ("Z", "Y"):
            indices = (2, 1)
        case _:
            raise ValueError
    return indices



"""> PyDRex: 2D corner flow tests."""

import contextlib as cl
import functools as ft
import pathlib as pl
from time import perf_counter

import numpy as np
import pytest

from pydrex import core as _core
from pydrex import diagnostics as _diagnostics
from pydrex import geometry as _geo
from pydrex import io as _io
from pydrex import logger as _log
from pydrex import minerals as _minerals
from pydrex import pathlines as _path
from pydrex import stats as _stats
from pydrex import utils as _utils
from pydrex import velocity as _velocity
from pydrex import visualisation as _vis

Pool, HAS_RAY = _utils.import_proc_pool()

# Subdirectory of `outdir` used to store outputs from these tests.
SUBDIR = "2d_cornerflow"


class TestOlivineA:
    """Tests for pure A-type olivine polycrystals in 2D corner flows."""

    class_id = "corner_olivineA"

    @classmethod
    def run(
        cls,
        params,
        seed,
        get_velocity,
        get_velocity_gradient,
        min_coords,
        max_coords,
        max_strain,
        n_timesteps,
        final_location,
    ):
        """Run 2D corner flow A-type olivine simulation."""
        mineral = _minerals.Mineral(
            _core.MineralPhase.olivine,
            _core.MineralFabric.olivine_A,
            _core.DeformationRegime.dislocation,
            n_grains=params["number_of_grains"],
            seed=seed,
        )
        deformation_gradient = np.eye(3)
        timestamps, get_position = _path.get_pathline(
            final_location,
            get_velocity,
            get_velocity_gradient,
            min_coords,
            max_coords,
            max_strain,
            regular_steps=n_timesteps,
        )
        positions = [get_position(t) for t in timestamps]
        velocity_gradients = [
            get_velocity_gradient(np.nan, np.asarray(x)) for x in positions
        ]
        strains = np.empty_like(timestamps)
        strains[0] = 0
        for t, time in enumerate(timestamps[:-1], start=1):
            strains[t] = strains[t - 1] + (
                _utils.strain_increment(timestamps[t] - time, velocity_gradients[t])
            )
            _log.info(
                "final location = %s; step %d/%d (ε = %.2f)",
                final_location.ravel(),
                t,
                len(timestamps) - 1,
                strains[t],
            )

            deformation_gradient = mineral.update_orientations(
                params,
                deformation_gradient,
                get_velocity_gradient,
                pathline=(time, timestamps[t], get_position),
            )
        return timestamps, positions, strains, mineral, deformation_gradient

    @pytest.mark.slow
    def test_steady4(self, outdir, seed, ncpus):
        """Test CPO evolution in steady 2D corner flow along 4 pathlines.

        Initial condition: random orientations and uniform volumes in all `Mineral`s.

        Plate velocity: 2 cm/yr

        .. note::
            This example takes about 11 CPU hours to run and uses around 60GB of RAM.
            It is recommended to only use `ncpus=4` which matches the number of
            pathlines, because higher numbers can lead to redundant cross-core
            communication.

        """
        # Plate speed (half spreading rate), convert cm/yr to m/s.
        plate_speed = 2.0 / (100.0 * 365.0 * 86400.0)
        domain_height = 2.0e5  # Represents the depth of olivine-spinel transition.
        domain_width = 1.0e6
        params = _io.DEFAULT_PARAMS
        params["number_of_grains"] = 5000
        n_timesteps = 50  # Number of places along the pathline to compute CPO.
        get_velocity, get_velocity_gradient = _velocity.corner_2d("X", "Z", plate_speed)
        # Z-values at the end of each pathline.
        z_ends = list(map(lambda x: x * domain_height, (-0.1, -0.3, -0.54, -0.78)))

        # Optional plotting and logging setup.
        optional_logging = cl.nullcontext()
        if outdir is not None:
            out_basepath = f"{outdir}/{SUBDIR}/{self.class_id}_prescribed"
            optional_logging = _log.logfile_enable(f"{out_basepath}.log")
            npzpath = pl.Path(f"{out_basepath}.npz")
            labels = []
            angles = []
            indices = []
            paths = []
            path_strains = []
            directions = []
            max_grainsizes = []
            timestamps = []

        _run = ft.partial(
            self.run,
            params,
            seed,
            get_velocity,
            get_velocity_gradient,
            np.array([0.0, 0.0, -domain_height]),
            np.array([domain_width, 0.0, 0.0]),
            10,
            n_timesteps,
        )

        final_locations = [np.array([domain_width, 0.0, z_exit]) for z_exit in z_ends]
        with optional_logging:
            _begin = perf_counter()
            with Pool(processes=ncpus) as pool:
                for i, out in enumerate(pool.imap(_run, final_locations)):
                    times, positions, strains, mineral, deformation_gradient = out
                    _log.info("final deformation gradient:\n%s", deformation_gradient)
                    z_exit = z_ends[i]
                    if outdir is not None:
                        mineral.save(npzpath, _io.stringify(z_exit))

                    misorient_angles = np.zeros(n_timesteps)
                    bingham_vectors = np.zeros((n_timesteps, 3))
                    orientations_resampled, f_resampled = _stats.resample_orientations(
                        mineral.orientations, mineral.fractions, seed=seed
                    )
                    misorient_indices = _diagnostics.misorientation_indices(
                        orientations_resampled,
                        _geo.LatticeSystem.orthorhombic,
                        pool=pool,
                    )
                    for idx, matrices in enumerate(orientations_resampled):
                        direction_mean = _diagnostics.bingham_average(
                            matrices,
                            axis=_minerals.OLIVINE_PRIMARY_AXIS[mineral.fabric],
                        )
                        misorient_angles[idx] = _diagnostics.smallest_angle(
                            direction_mean,
                            np.asarray([1, 0, 0], dtype=np.float64),
                        )
                        bingham_vectors[idx] = direction_mean

                    _log.debug("Total walltime: %s", perf_counter() - _begin)

                    if outdir is not None:
                        labels.append(rf"$z_{{f}}$ = {z_exit/1e3:.1f} km")
                        angles.append(misorient_angles)
                        indices.append(misorient_indices)
                        paths.append(positions)
                        max_grainsizes.append(
                            np.asarray([np.max(f) for f in f_resampled])
                        )
                        path_strains.append(strains)
                        timestamps.append(times)
                        directions.append(bingham_vectors)

        if outdir is not None:
            np.savez(
                f"{out_basepath}_path.npz",
                strains_1=path_strains[0],
                strains_2=path_strains[1],
                strains_3=path_strains[2],
                strains_4=path_strains[3],
                paths_1=np.asarray(paths[0]),
                paths_2=np.asarray(paths[1]),
                paths_3=np.asarray(paths[2]),
                paths_4=np.asarray(paths[3]),
            )
            np.savez(
                f"{out_basepath}_strength.npz",
                strength_1=indices[0],
                strength_2=indices[1],
                strength_3=indices[2],
                strength_4=indices[3],
            )
            np.savez(
                f"{out_basepath}_angles.npz",
                angles_1=angles[0],
                angles_2=angles[1],
                angles_3=angles[2],
                angles_4=angles[3],
            )
            markers = ("s", "o", "v", "*")
            cmaps = ["cmc.batlow_r"] * len(markers)
            fig_domain = _vis.figure(figsize=(10, 3))
            ax_domain = fig_domain.add_subplot()
            for i, z_exit in enumerate(z_ends):
                if i == 0:
                    resolution = [25, 5]
                else:
                    resolution = None
                _vis.pathline_box2d(
                    ax_domain,
                    get_velocity,
                    "xz",
                    path_strains[i],
                    paths[i],
                    markers[i],
                    [0, -domain_height],
                    [domain_width + 1e5, 0],
                    resolution,
                    cmap=cmaps[i],
                    cpo_vectors=directions[i],
                    cpo_strengths=indices[i],
                    scale=25,
                    scale_units="width",
                )

            fig_strength, ax_strength, colors = _vis.strengths(
                None,
                path_strains,
                indices,
                "CPO strength (M-index)",
                markers,
                labels,
                colors=path_strains,
                cmaps=cmaps,
            )
            fig_alignment, ax_alignment, colors = _vis.alignment(
                None,
                path_strains,
                angles,
                markers,
                labels,
                colors=path_strains,
                cmaps=cmaps,
            )

            fig_domain.savefig(_io.resolve_path(f"{out_basepath}_path.pdf"))
            fig_strength.savefig(_io.resolve_path(f"{out_basepath}_strength.pdf"))
            fig_alignment.savefig(_io.resolve_path(f"{out_basepath}_angles.pdf"))



"""Find seeds that make grains start with an average SCCS direction of 45."""

import numpy as np

from pydrex import diagnostics as _d
from pydrex import logger as _l
from pydrex import minerals as _m

for i in range(10000):
    with _l.handler_level("ERROR"):
        m = _m.Mineral(seed=i, n_grains=4394)
        if (
            np.abs(
                _d.smallest_angle(
                    _d.elasticity_components(
                        _m.voigt_averages([m], {"olivine_fraction": 1})[0]
                    )["hexagonal_axis"][0],
                    np.asarray([1, 0, 0], dtype=np.float64),
                )
                - 45
            )
            < 0.1
        ):
            print(i)



"""> PyDRex: Entry points and argument handling for command line tools.

All CLI handlers should be registered in the `CLI_HANDLERS` namedtuple,
which ensures that they will be installed as executable scripts alongside the package.

"""

import argparse
import os
from collections import namedtuple
from zipfile import ZipFile

from pydrex import exceptions as _err
from pydrex import io as _io
from pydrex import logger as _log
from pydrex import minerals as _minerals
from pydrex import stats as _stats
from pydrex import visualisation as _vis


class CliTool:
    """Base class for CLI tools defining the required interface."""

    def __call__(self):
        return NotImplementedError

    def _get_args(self) -> argparse.Namespace:
        return NotImplementedError


class MeshGenerator(CliTool):
    """PyDRex script to generate various simple meshes.

    Only rectangular (2D) meshes are currently supported. The RESOLUTION must be a comma
    delimited set of directives of the form `<LOC>:<RES>` where `<LOC>` is a location
    specifier, i.e. either "G" (global) or a compas direction like "N", "S", "NE", etc.,
    and `<RES>` is a floating point value to be set as the resolution at that location.

    """

    def __call__(self):
        try:  # This one is dangerous, especially in CI.
            from pydrex import mesh as _mesh
        except ImportError:
            raise _err.MissingDependencyError(
                "missing optional meshing dependencies."
                + " Have you installed the package with 'pip install pydrex[mesh]'?"
            )

        args = self._get_args()

        if args.kind == "rectangle":
            if args.center is None:
                center = (0, 0)
            else:
                center = [float(s) for s in args.center.split(",")]
                assert len(center) == 2

            width, height = map(float, args.size.split(","))
            _loc_map = {
                "G": "global",
                "N": "north",
                "S": "south",
                "E": "east",
                "W": "west",
                "NE": "north-east",
                "NW": "north-west",
                "SE": "south-east",
                "SW": "south-west",
            }
            try:
                resolution = {
                    _loc_map[k]: float(v)
                    for k, v in map(lambda s: s.split(":"), args.resolution.split(","))
                }
            except KeyError:
                raise KeyError(
                    "invalid or unsupported location specified in resolution directive"
                ) from None
            except ValueError:
                raise ValueError(
                    "invalid resolution value. The format should be '<LOC1>:<RES1>,<LOC2>:<RES2>,...'"
                ) from None
            _mesh.rectangle(
                args.output[:-4],
                (args.ref_axes[0], args.ref_axes[1]),
                center,
                width,
                height,
                resolution,
            )

    def _get_args(self) -> argparse.Namespace:
        description, epilog = self.__doc__.split(os.linesep + os.linesep, 1)
        parser = argparse.ArgumentParser(description=description, epilog=epilog)
        parser.add_argument("size", help="width,height[,depth] of the mesh")
        parser.add_argument(
            "-r",
            "--resolution",
            help="resolution for the mesh (edge length hint(s) for gmsh)",
            required=True,
        )
        parser.add_argument("output", help="output file (.msh)")
        parser.add_argument(
            "-c",
            "--center",
            help="center of the mesh as 2 or 3 comma-separated coordinates. default: (0, 0[, 0])",
            default=None,
        )
        parser.add_argument(
            "-a",
            "--ref-axes",
            help=(
                "two letters from {'x', 'y', 'z'} that specify"
                + " the horizontal and vertical axes of the mesh"
            ),
            default="xz",
        )
        parser.add_argument(
            "-k", "--kind", help="kind of mesh, e.g. 'rectangle'", default="rectangle"
        )
        return parser.parse_args()


class H5partExtractor(CliTool):
    """PyDRex script to extract raw CPO data from Fluidity .h5part files.

    Fluidity saves data stored on model `particles` to an `.h5part` file.
    This script converts that file to canonical serialisation formats:
    - a `.npz` file containing the raw CPO orientations and (surrogate) grain sizes
    - an `.scsv` file containing the pathline positions and accumulated strain

    It is assumed that CPO data is stored in keys called 'CPO_<N>' in the .h5part
    data, where `<N>` is an integer in the range 1—`n_grains`. The accumulated strain is
    read from the attribute `CPO_<S>` where S=`ngrains`+1. Particle positions are read
    from the attributes `x`, `y`, and `z`.

    At the moment, dynamic changes in fabric or phase are not supported.

    """

    def __call__(self):
        args = self._get_args()
        _io.extract_h5part(
            args.input, args.phase, args.fabric, args.ngrains, args.output
        )

    def _get_args(self) -> argparse.Namespace:
        description, epilog = self.__doc__.split(os.linesep + os.linesep, 1)
        parser = argparse.ArgumentParser(description=description, epilog=epilog)
        parser.add_argument("input", help="input file (.h5part)")
        parser.add_argument(
            "-p",
            "--phase",
            help="type of `pydrex.MineralPhase` (as an ordinal number); 0 by default",
            default=0,
        )
        parser.add_argument(
            "-f",
            "--fabric",
            type=int,
            help="type of `pydrex.MineralFabric` (as an ordinal number); 0 by default",
            default=0,
        )
        parser.add_argument(
            "-n",
            "--ngrains",
            help="number of grains used in the Fluidity simulation",
            type=int,
            required=True,
        )
        parser.add_argument(
            "-o",
            "--output",
            help="filename for the output NPZ file (stem also used for the .scsv)",
            required=True,
        )
        return parser.parse_args()


class NPZFileInspector(CliTool):
    """PyDRex script to show information about serialized CPO data.

    Lists the keys that should be used for the `postfix` in `pydrex.Mineral.load` and
    `pydrex.Mineral.from_file`.

    """

    def __call__(self):
        args = self._get_args()
        with ZipFile(args.input) as npz:
            names = npz.namelist()
            print("NPZ file with keys:")
            for name in names:
                if not (
                    name.startswith("meta")
                    or name.startswith("fractions")
                    or name.startswith("orientations")
                ):
                    _log.warning(f"found unknown NPZ key '{name}' in '{args.input}'")
                print(f" - {name}")

    def _get_args(self) -> argparse.Namespace:
        description, epilog = self.__doc__.split(os.linesep + os.linesep, 1)
        parser = argparse.ArgumentParser(description=description, epilog=epilog)
        parser.add_argument("input", help="input file (.npz)")
        return parser.parse_args()


class PoleFigureVisualiser(CliTool):
    """PyDRex script to plot pole figures of serialized CPO data.

    Produces [100], [010] and [001] pole figures for serialized `pydrex.Mineral`s.
    If the range of indices is not specified,
    a maximum of 25 of each pole figure will be produced by default.

    """

    def __call__(self):
        try:
            args = self._get_args()
            if args.range is None:
                i_range = None
            else:
                start, stop_ex, step = (int(s) for s in args.range.split(":"))
                # Make command line start:stop:step stop-inclusive, it's more intuitive.
                i_range = range(start, stop_ex + step, step)

            density_kwargs = {"kernel": args.kernel}
            if args.smoothing is not None:
                density_kwargs["σ"] = args.smoothing

            mineral = _minerals.Mineral.from_file(args.input, postfix=args.postfix)
            if i_range is None:
                i_range = range(0, len(mineral.orientations))
                if len(i_range) > 25:
                    _log.warning(
                        "truncating to 25 timesteps (out of %s total)", len(i_range)
                    )
                    i_range = range(0, 25)

            orientations_resampled, _ = _stats.resample_orientations(
                mineral.orientations[i_range.start : i_range.stop : i_range.step],
                mineral.fractions[i_range.start : i_range.stop : i_range.step],
            )
            if args.scsv is None:
                strains = None
            else:
                strains = _io.read_scsv(args.scsv).strain[
                    i_range.start : i_range.stop : i_range.step
                ]
            _vis.polefigures(
                orientations_resampled,
                ref_axes=args.ref_axes,
                i_range=i_range,
                density=args.density,
                savefile=args.out,
                strains=strains,
                **density_kwargs,
            )
        except (argparse.ArgumentError, ValueError, _err.Error) as e:
            _log.error(str(e))

    def _get_args(self) -> argparse.Namespace:
        description, epilog = self.__doc__.split(os.linesep + os.linesep, 1)
        parser = argparse.ArgumentParser(description=description, epilog=epilog)
        parser.add_argument("input", help="input file (.npz)")
        parser.add_argument(
            "-r",
            "--range",
            help="range of strain indices to be plotted, in the format start:stop:step",
            default=None,
        )
        parser.add_argument(
            "-f",
            "--scsv",
            help=(
                "path to SCSV file with a column named 'strain'"
                + " that lists shear strain percentages for each strain index"
            ),
            default=None,
        )
        parser.add_argument(
            "-p",
            "--postfix",
            help=(
                "postfix of the mineral to load,"
                + " required if the input file contains data for multiple minerals"
            ),
            default=None,
        )
        parser.add_argument(
            "-d",
            "--density",
            help="toggle contouring of pole figures using point density estimation",
            default=False,
            action="store_true",
        )
        parser.add_argument(
            "-k",
            "--kernel",
            help=(
                "kernel function for point density estimation, one of:"
                + f" {list(_stats.SPHERICAL_COUNTING_KERNELS.keys())}"
            ),
            default="linear_inverse_kamb",
        )
        parser.add_argument(
            "-s",
            "--smoothing",
            help="smoothing parameter for Kamb type density estimation kernels",
            default=None,
            type=float,
            metavar="σ",
        )
        parser.add_argument(
            "-a",
            "--ref-axes",
            help=(
                "two letters from {'x', 'y', 'z'} that specify"
                + " the horizontal and vertical axes of the pole figures"
            ),
            default="xz",
        )
        parser.add_argument(
            "-o",
            "--out",
            help="name of the output file, with either .png or .pdf extension",
            default="polefigures.png",
        )
        return parser.parse_args()


# These are not the final names of the executables (those are set in pyproject.toml).
_CLI_HANDLERS = namedtuple(
    "CLI_HANDLERS",
    {
        "pole_figure_visualiser",
        "npz_file_inspector",
        "mesh_generator",
        "h5part_extractor",
    },
)
CLI_HANDLERS = _CLI_HANDLERS(
    pole_figure_visualiser=PoleFigureVisualiser(),
    npz_file_inspector=NPZFileInspector(),
    mesh_generator=MeshGenerator(),
    h5part_extractor=H5partExtractor(),
)



#!/usr/bin/env python3
from numpy import amax, amin, array, diag, hstack, inf, linspace

name = "Tag"  # A little tag to keep track

dim = 2  # Number of dimensions; must be 2 or 3
assert dim in [2, 3]

# Generate a regular grid
uniformGridSpacing = True
if uniformGridSpacing:  # As many entries as dimensions: X, [Y], Z
    gridRes = array([4e3, 4e3])
    gridMin = array([0, 0])
    gridMax = array([1.2e6, 4e5]) + gridRes
    gridNodes = ((gridMax - gridMin) / gridRes + 1).astype(int)
    gridCoords = [linspace(x, y, z) for x, y, z in zip(gridMin, gridMax, gridNodes)]
else:  # As many variables as dimensions: X, [Y], Z
    xCoords = hstack(
        (
            linspace(5e5, 1.3e6, 161),
            linspace(1.3025e6, 1.7e6, 160),
            linspace(1.705e6, 2.5e6, 160),
        )
    )
    yCoords = hstack(
        (
            linspace(0, 3e5, 13),
            linspace(3.2e5, 4e5, 5),
            linspace(4.1e5, 4.5e5, 5),
            linspace(4.55e5, 4.8e5, 6),
            linspace(4.825e5, 5.2e5, 16),
            linspace(5.25e5, 5.5e5, 6),
            linspace(5.6e5, 6e5, 5),
            linspace(6.2e5, 7e5, 5),
            linspace(7.25e5, 1e6, 12),
        )
    )
    zCoords = hstack(
        (
            linspace(0, 1e5, 51),
            linspace(1.04e5, 2e5, 25),
            linspace(2.1e5, 3e5, 10),
            linspace(3.2e5, 4e5, 5),
        )
    )
    gridCoords = [xCoords, yCoords, zCoords]
    gridNodes = [arr.size for arr in gridCoords]
    gridMin = [amin(arr) for arr in gridCoords]
    gridMax = [amax(arr) for arr in gridCoords]
assert dim == len(gridCoords)

checkpoint = 3000  # Number of nodes after which a checkpoint should be dumped

Xol = 0.7  # Fraction of olivine in the aggregate
tau = array([1, 2, 3, inf])  # Olivine
tau_ens = 1  # Enstatite
mob = 125  # Grain boundary mobility
chi = 0.3  # Threshold volume fraction for grain boundary sliding
lamb = 5  # Nucleation parameter set to 5 as in Kaminski and Ribe (2001)
size = 13**3  # Initial size | Number of points in the Eulerian space
stressexp = 3.5  # Stress exponent for olivine and enstatite
# Stiffness matrix for Olivine (GPa)
S0 = diag((320.71, 197.25, 234.32, 63.77, 77.67, 78.36))
S0[0, 1] = S0[1, 0] = 69.84
S0[0, 2] = S0[2, 0] = 71.22
S0[1, 2] = S0[2, 1] = 74.8
# Stiffness matrix for Enstatite (GPa)
S0_ens = diag((236.9, 180.5, 230.4, 84.3, 79.4, 80.1))
S0_ens[0, 1] = S0_ens[1, 0] = 79.6
S0_ens[0, 2] = S0_ens[2, 0] = 63.2
S0_ens[1, 2] = S0_ens[2, 1] = 56.8



"""> PyDRex: Computations of mineral texture and elasticity.

**Acronyms:**
- CPO = Crystallographic preferred orientation,
    i.e. preferential clustering of polycrystal grain orientations in SO(3),
    leading to an overall anisotropic orientation distribution

"""

import io
from dataclasses import dataclass, field
from zipfile import ZipFile

import numpy as np
from numpy import asarray as Ŋ
from scipy import linalg as la
from scipy.integrate import LSODA
from scipy.spatial.transform import Rotation

from pydrex import core as _core
from pydrex import exceptions as _err
from pydrex import io as _io
from pydrex import logger as _log
from pydrex import tensors as _tensors
from pydrex import utils as _utils

OLIVINE_STIFFNESS = np.array(
    [
        [320.71, 69.84, 71.22, 0.0, 0.0, 0.0],
        [69.84, 197.25, 74.8, 0.0, 0.0, 0.0],
        [71.22, 74.8, 234.32, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 63.77, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.0, 77.67, 0.0],
        [0.0, 0.0, 0.0, 0.0, 0.0, 78.36],
    ]
)
"""Stiffness tensor for olivine (Voigt representation), with units of GPa.

The source of the values used here is unknown, but they are copied
from the original DRex code: <http://www.ipgp.fr/~kaminski/web_doudoud/DRex.tar.gz>
[88K download]

"""


ENSTATITE_STIFFNESS = np.array(
    [
        [236.9, 79.6, 63.2, 0.0, 0.0, 0.0],
        [79.6, 180.5, 56.8, 0.0, 0.0, 0.0],
        [63.2, 56.8, 230.4, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 84.3, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.0, 79.4, 0.0],
        [0.0, 0.0, 0.0, 0.0, 0.0, 80.1],
    ]
)
"""Stiffness tensor for enstatite (Voigt representation), with units of GPa.

The source of the values used here is unknown, but they are copied
from the original DRex code: <http://www.ipgp.fr/~kaminski/web_doudoud/DRex.tar.gz>
[88K download]

"""


OLIVINE_PRIMARY_AXIS = {
    _core.MineralFabric.olivine_A: "a",
    _core.MineralFabric.olivine_B: "c",
    _core.MineralFabric.olivine_C: "c",
    _core.MineralFabric.olivine_D: "a",
    _core.MineralFabric.olivine_E: "a",
}
"""Primary slip axis name for for the given olivine `fabric`."""


OLIVINE_SLIP_SYSTEMS = (
    ([0, 1, 0], [1, 0, 0]),
    ([0, 0, 1], [1, 0, 0]),
    ([0, 1, 0], [0, 0, 1]),
    ([1, 0, 0], [0, 0, 1]),
)
"""Slip systems for olivine in conventional order.

Tuples contain the slip plane normal and slip direction vectors.
The order of slip systems returned matches the order of critical shear stresses
returned by `pydrex.core.get_crss`.

"""


# TODO: Compare to [Man & Huang, 2011](https://doi.org/10.1007/s10659-011-9312-y).
def voigt_averages(minerals, weights):
    """Calculate elastic tensors as the Voigt averages of a collection of `mineral`s.

    Args:
    - `minerals` — list of `pydrex.minerals.Mineral` instances storing orientations and
      fractional volumes of the grains within each distinct mineral phase
    - `weights` (dict) — dictionary containing weights of each mineral
      phase, as a fraction of 1, in keys named "<phase>_fraction",
      e.g. "olivine_fraction"

    Raises a ValueError if the minerals contain an unequal number of grains or stored
    texture results.

    """
    n_grains = minerals[0].n_grains
    if not np.all([m.n_grains == n_grains for m in minerals[1:]]):
        raise ValueError("cannot average minerals with unequal grain counts")
    n_steps = len(minerals[0].orientations)
    if not np.all([len(m.orientations) == n_steps for m in minerals[1:]]):
        raise ValueError(
            "cannot average minerals with variable-length orientation arrays"
        )
    if not np.all([len(m.fractions) == n_steps for m in minerals]):
        raise ValueError(
            "cannot average minerals with variable-length grain volume arrays"
        )

    elastic_tensors = {}

    # TODO: Perform rotation directly on the 6x6 matrices, see Carcione 2007.
    # This trick is implemented in cpo_elastic_tensor.cc in Aspect.
    average_tensors = np.zeros((n_steps, 6, 6))
    for i in range(n_steps):
        for mineral in minerals:
            for n in range(n_grains):
                match mineral.phase:
                    case _core.MineralPhase.olivine:
                        if "olivine" not in elastic_tensors:
                            elastic_tensors["olivine"] = (
                                _tensors.voigt_to_elastic_tensor(OLIVINE_STIFFNESS)
                            )
                        average_tensors[i] += _tensors.elastic_tensor_to_voigt(
                            _tensors.rotate(
                                elastic_tensors["olivine"],
                                mineral.orientations[i][n, ...].transpose(),
                            )
                            * mineral.fractions[i][n]
                            * weights["olivine_fraction"]
                        )
                    case _core.MineralPhase.enstatite:
                        if "enstatite" not in elastic_tensors:
                            elastic_tensors["enstatite"] = (
                                _tensors.voigt_to_elastic_tensor(ENSTATITE_STIFFNESS)
                            )
                        average_tensors[i] += _tensors.elastic_tensor_to_voigt(
                            _tensors.rotate(
                                elastic_tensors["enstatite"],
                                minerals.orientations[i][n, ...].transpose(),
                            )
                            * mineral.fractions[i][n]
                            * weights["enstatite_fraction"]
                        )
                    case _:
                        raise ValueError(f"unsupported mineral phase: {mineral.phase}")
    return average_tensors


@dataclass
class Mineral:
    """Class for storing polycrystal texture for a single mineral phase.

    A `Mineral` stores texture data for an aggregate of grains*.
    Additionally, mineral fabric type and deformation regime are also tracked.
    To provide an initial texture for the mineral, use the constructor arguments
    `fractions_init` and `orientations_init`. By default,
    a uniform volume distribution of random orientations is generated.

    The `update_orientations` method computes new orientations and grain volumes
    for a given velocity gradient. These results are stored in the `.orientations` and
    `.fractions` attributes of the `Mineral` instance. The method also returns the
    updated macroscopic deformation gradient based on the provided initial deformation
    gradient.

    *Note that the "number of grains" is a static integer value that
    does not track the actual number of physical grains in the deforming polycrystal.
    Instead, this number acts as a "number of bins" for the statistical resolution of
    the crystallographic orientation distribution. The value is roughly equivalent to
    (a multiple of) the number of initial, un-recrystallised grains in the polycrystal.
    It is assumed that recrystallised grains do not grow large enough to require
    rotation tracking.

    **Examples:**

    Mineral with isotropic initial texture:

    >>> import pydrex
    >>> olA = pydrex.Mineral(
    ...     phase=pydrex.MineralPhase.olivine,
    ...     fabric=pydrex.MineralFabric.olivine_A,
    ...     regime=pydrex.DeformationRegime.dislocation,
    ...     n_grains=2000
    ... )
    >>> olA.phase
    <MineralPhase.olivine: 0>
    >>> olA.fabric
    <MineralFabric.olivine_A: 0>
    >>> olA.regime
    <DeformationRegime.dislocation: 1>
    >>> olA.n_grains
    2000

    Mineral with specified initial texture and default phase, fabric and regime settings
    which are for an olivine A-type mineral in the dislocation creep regime.
    The initial grain volume fractions should be normalised.

    >>> import numpy as np
    >>> from scipy.spatial.transform import Rotation
    >>> import pydrex
    >>> rng = np.random.default_rng()
    >>> n_grains = 2000
    >>> olA = pydrex.Mineral(
    ...     n_grains=n_grains,
    ...     fractions_init=np.full(n_grains, 1 / n_grains),
    ...     orientations_init=Rotation.from_euler(
    ...         "zxz", [
    ...             [x * np.pi / 2, np.pi / 2, np.pi / 2] for x in rng.random(n_grains)
    ...         ]
    ...     ).inv().as_matrix(),
    ... )
    >>> len(olA.orientations)
    1
    >>> type(olA.orientations)
    <class 'list'>
    >>> olA.orientations[0].shape
    (2000, 3, 3)
    >>> olA.fractions[0].shape
    (2000,)

    Note that minerals can also be constructed from serialized data,
    see `Mineral.load` and `Mineral.from_file`.

    **Attributes:**
    - `phase` (`pydrex.core.MineralPhase`) — ordinal number of the mineral phase
    - `fabric` (`pydrex.core.MineralFabric`) — ordinal number of the fabric type
    - `regime` (`pydrex.core.DeformationRegime`) — ordinal number of the deformation
      regime
    - `n_grains` (int) — number of grains in the aggregate
    - `fractions` (list of arrays) — grain volume fractions for each texture snapshot
    - `orientations` (list of arrays) — grain orientation matrices for each texture
      snapshot
    - `seed` (int) — seed used by the random number generator to set up the isotropic
      initial condition when `fractions_init` or `orientations_init` are not provided
    - `lband` (int) — passed to the `scipy.integrate.LSODA` solver
    - `uband` (int) — passed to the `scipy.integrate.LSODA` solver

    """

    phase: int = _core.MineralPhase.olivine
    fabric: int = _core.MineralFabric.olivine_A
    regime: int = _core.DeformationRegime.dislocation
    n_grains: int = _io.DEFAULT_PARAMS["number_of_grains"]
    # Initial condition, randomised if not given.
    fractions_init: np.ndarray = None
    orientations_init: np.ndarray = None
    fractions: list = field(default_factory=list)
    orientations: list = field(default_factory=list)
    seed: int = None
    lband: int = None
    uband: int = None

    def __str__(self):
        # String output, used for str(self) and f"{self}", etc.
        if hasattr(self.fractions[0], "shape"):
            shape_of_fractions = str(self.fractions[0].shape)
        else:
            shape_of_fractions = "(?)"

        if hasattr(self.orientations[0], "shape"):
            shape_of_orientations = str(self.orientations[0].shape)
        else:
            shape_of_orientations = "(?)"

        obj = self.__class__.__qualname__
        phase = f"(phase={self.phase!r}, "
        fabric = f"fabric={self.fabric!r}, "
        regime = f"regime={self.regime!r}, "
        n_grains = f"n_grains={self.n_grains}, "
        _fclass = self.fractions.__class__.__qualname__
        _f0class = self.fractions[0].__class__.__qualname__
        frac = f"fractions=<{_fclass} of {_f0class} {shape_of_fractions}>, "
        _oclass = self.orientations.__class__.__qualname__
        _o0class = self.orientations[0].__class__.__qualname__
        orient = f"orientations=<{_oclass} of {_o0class} {shape_of_orientations}>)"
        return f"{obj}{phase}{fabric}{regime}{n_grains}{frac}{orient}"

    def _repr_pretty_(self, p, cycle):
        # Format to use when printing to IPython or other interactive console.
        p.text(self.__str__() if not cycle else self.__class__.__qualname__ + "(...)")

    def __post_init__(self):
        """Initialise random orientations and grain volume fractions."""
        if self.fractions_init is None:
            self.fractions_init = np.full(self.n_grains, 1.0 / self.n_grains)
        if self.orientations_init is None:
            self.orientations_init = Rotation.random(
                self.n_grains, random_state=self.seed
            ).as_matrix()
        # For large numbers of grains, the number of ODE's exceeds what LSODA can
        # handle. Therefore, we specify the Jacobian matrix as banded.
        # By default, we use a bandwidth o f 12000 with lband = uband = 6000.
        # This should work for up to 10000 grains.
        if self.lband is None and self.uband is None and self.n_grains > 4632:
            _log.warning(
                "using a banded Jacobian because of the large number of grains."
                + " To manually control the bandwidth, set `lband` and/or `uband`"
                + f" in calls to `{self.__class__.__qualname__}.update_orientations`."
            )
            self.lband = 6000
            self.uband = 6000

        # Copy the initial values to the storage lists.
        self.fractions.append(self.fractions_init)
        self.orientations.append(self.orientations_init)

        # Delete the initial value duplicates to avoid confusion.
        del self.fractions_init
        del self.orientations_init

        _log.info("created %s", self)

    def __eq__(self, other):
        if other.__class__ is self.__class__:
            return (
                self.phase == other.phase
                and self.fabric == other.fabric
                and self.regime == other.regime
                and self.n_grains == other.n_grains
                and len(self.fractions) == len(other.fractions)
                and np.all(
                    Ŋ([f.shape for f in self.fractions])
                    == Ŋ([f.shape for f in other.fractions])
                )
                and np.all(Ŋ(self.fractions) == Ŋ(other.fractions))
                and len(self.orientations) == len(other.orientations)
                and np.all(
                    Ŋ([f.shape for f in self.orientations])
                    == Ŋ([f.shape for f in other.orientations])
                )
                and np.all(Ŋ(self.orientations) == Ŋ(other.orientations))
                and self.seed == other.seed
                and self.lband == other.lband
                and self.uband == other.uband
            )
        return False

    def update_orientations(
        self,
        config,
        deformation_gradient,
        get_velocity_gradient,
        pathline,
        **kwargs,
    ):
        """Update orientations and volume distribution for the `Mineral`.

        Update crystalline orientations and grain volume distribution
        for minerals undergoing plastic deformation. Return the updated deformation
        gradient measuring the corresponding macroscopic deformation.

        Args:
        - `config` (dict) — PyDRex configuration dictionary
        - `deformation_gradient` (array) — 3x3 deformation gradient tensor
        - `get_velocity_gradient` (function) — callable with signature f(t, x) that
          returns a 3x3 velocity gradient matrix at time t and position x (3D vector)
        - `pathline` (tuple) — tuple consisting of:
            1. the time at which to start the CPO integration (t_start)
            2. the time at which to stop the CPO integration (t_end)
            3. callable with signature f(t) that returns the position of the mineral at
                time t ∈ [t_start, t_end]

        Any additional (optional) keyword arguments are passed to
        `scipy.integrate.LSODA`.

        Array values must provide a NumPy-compatible interface:
        <https://numpy.org/doc/stable/user/whatisnumpy.html>

        Examples:

        >>> from itertools import pairwise
        >>> import pydrex
        >>> olA = pydrex.Mineral(
        ...           phase=pydrex.MineralPhase.olivine,
        ...           fabric=pydrex.MineralFabric.olivine_A,
        ...           regime=pydrex.DeformationRegime.dislocation,
        ...           n_grains=pydrex.DEFAULT_PARAMS["number_of_grains"],
        ... )
        >>> def get_velocity_gradient(t, x):  # Simple L for simple shear.
        ...     L = np.zeros((3, 3))
        ...     L[0, 2] = 1
        ...     return L
        >>> def get_position(t):
        ...     return np.zeros(3)  # Stationary polycrystal for this example.
        >>> timestamps = range(2)  # Just 1 timestep for demonstration.
        >>> deformation_gradient = np.eye(3)  # Start with an undeformed polycrystal.
        >>> for t_start, t_end in pairwise(timestamps):
        ...     # Update deformation_gradient, olA.orientations and olA.fractions.
        ...     deformation_gradient = olA.update_orientations(
        ...         pydrex.DEFAULT_PARAMS,
        ...         deformation_gradient,
        ...         get_velocity_gradient,
        ...         (t_start, t_end, get_position),
        ...     )
        >>> np.all(deformation_gradient == np.eye(3) + get_velocity_gradient(t_end, np.nan))
        True

        """

        # ===== Set up callables for the ODE solver and internal processing =====

        def eval_rhs(t, y):
            """Evaluate right hand side of the D-Rex PDE."""
            # assert not np.any(np.isnan(y)), y[np.isnan(y)].shape
            position = get_position(t)
            velocity_gradient = get_velocity_gradient(t, position)
            # _log.debug(
            #     "calculating CPO at %s (t=%e) with velocity gradient %s",
            #     position,
            #     t,
            #     velocity_gradient.ravel(),
            # )

            if self.phase == _core.MineralPhase.olivine:
                volume_fraction = config["olivine_fraction"]
            elif self.phase == _core.MineralPhase.enstatite:
                volume_fraction = config["enstatite_fraction"]
            else:
                raise ValueError(
                    f"phase must be a valid `MineralPhase`, not {self.phase}"
                )

            strain_rate = (velocity_gradient + velocity_gradient.transpose()) / 2
            strain_rate_max = np.abs(la.eigvalsh(strain_rate)).max()
            deformation_gradient, orientations, fractions = _utils.extract_vars(
                y, self.n_grains
            )
            # Uses nondimensional values of strain rate and velocity gradient.
            orientations_diff, fractions_diff = _core.derivatives(
                phase=self.phase,
                fabric=self.fabric,
                n_grains=self.n_grains,
                orientations=orientations,
                fractions=fractions,
                strain_rate=strain_rate / strain_rate_max,
                velocity_gradient=velocity_gradient / strain_rate_max,
                stress_exponent=config["stress_exponent"],
                deformation_exponent=config["deformation_exponent"],
                nucleation_efficiency=config["nucleation_efficiency"],
                gbm_mobility=config["gbm_mobility"],
                volume_fraction=volume_fraction,
            )
            return np.hstack(
                (
                    (velocity_gradient @ deformation_gradient).flatten(),
                    orientations_diff.flatten() * strain_rate_max,
                    fractions_diff * strain_rate_max,
                )
            )

        def perform_step(solver):
            """Perform SciPy solver step and appropriate processing."""
            message = solver.step()
            if message is not None and solver.status == "failed":
                raise _err.IterationError(message)
            # _log.debug(
            #     "%s step_size=%e", solver.__class__.__qualname__, solver.step_size
            # )

            deformation_gradient, orientations, fractions = _utils.extract_vars(
                solver.y, self.n_grains
            )
            orientations, fractions = _utils.apply_gbs(
                orientations,
                fractions,
                config["gbs_threshold"],
                self.orientations[-1],
                self.n_grains,
            )
            solver.y[9:] = np.hstack((orientations.flatten(), fractions))

        # ===== Initialise and run the solver using the above callables =====

        time_start, time_end, get_position = pathline
        if not callable(get_velocity_gradient):
            raise ValueError(
                "unable to evaluate velocity gradient callable."
                + " You must provide a callable with signature f(t, x)"
                + " that returns a 3x3 matrix."
            )
        if not callable(get_position):
            raise ValueError(
                "unable to evaluate position callable."
                + " You must provide a callable with signature f(t)"
                + " that returns a 3-component array."
            )
        _log.debug(
            "calculating CPO from %s (t=%s) to %s (t=%s)",
            get_position(time_start),
            time_start,
            get_position(time_end),
            time_end,
        )
        _log.debug("    with deformation gradient %s", deformation_gradient.ravel())
        _log.debug(
            "    with velocity gradient interpolated between %s and %s",
            get_velocity_gradient(time_start, get_position(time_start)).ravel(),
            get_velocity_gradient(time_end, get_position(time_end)).ravel()
        )
        _log.debug(
            "    intermediate velocity gradient = %s",
            get_velocity_gradient(
                (time_start + time_end) / 2, get_position((time_start + time_end) / 2)
            ).ravel(),
        )

        y_start = np.hstack(
            (
                deformation_gradient.flatten(),
                self.orientations[-1].flatten(),
                self.fractions[-1],
            )
        )
        solver = LSODA(
            eval_rhs,
            time_start,
            y_start,
            time_end,
            atol=kwargs.pop("atol", np.abs(y_start * 1e-6) + 1e-12),
            rtol=kwargs.pop("rtol", 1e-6),
            first_step=kwargs.pop("first_step", np.abs(time_end - time_start) * 1e-1),
            # max_step=kwargs.pop("max_step", np.abs(time_end - time_start)),
            lband=self.lband,
            uband=self.uband,
            **kwargs,
        )
        perform_step(solver)
        while solver.status == "running":
            perform_step(solver)

        # Extract final values for this simulation step, append to storage.
        deformation_gradient, orientations, fractions = _utils.extract_vars(
            solver.y.squeeze(), self.n_grains
        )
        self.orientations.append(orientations)
        self.fractions.append(fractions)
        return deformation_gradient

    def save(self, filename, postfix=None):
        """Save CPO data for all stored timesteps to a `numpy` NPZ file.

        If `postfix` is not `None`, the data is appended to the NPZ file
        in fields ending with "`_postfix`".

        Raises a `ValueError` if the data shapes are not compatible.

        See also: `numpy.savez`, `Mineral.load`, `Mineral.from_file`.

        """
        if len(self.fractions) != len(self.orientations):
            raise ValueError(
                "Length of stored results must match."
                + " You've supplied currupted data with:\n"
                + f"- {len(self.fractions)} grain size results, and\n"
                + f"- {len(self.orientations)} orientation results."
            )
        if self.fractions[0].shape[0] == self.orientations[0].shape[0] == self.n_grains:
            data = {
                "meta": np.array(
                    [self.phase, self.fabric, self.regime], dtype=np.uint8
                ),
                "fractions": np.stack(self.fractions),
                "orientations": np.stack(self.orientations),
            }
            # Create parent directories, resolve relative paths.
            _io.resolve_path(filename)
            # Append to file, requires postfix (unique name).
            if postfix is not None:
                _log.info("saving Mineral to file %s (postfix: %s)", filename, postfix)
                archive = ZipFile(filename, mode="a", allowZip64=True)
                for key in data.keys():
                    with archive.open(
                        f"{key}_{postfix}", "w", force_zip64=True
                    ) as file:
                        buffer = io.BytesIO()
                        np.save(buffer, data[key])
                        file.write(buffer.getvalue())
                        buffer.close()
            else:
                _log.info("saving Mineral to file %s", filename)
                np.savez(filename, **data)
        else:
            raise ValueError(
                "Size of CPO data arrays must match number of grains."
                + " You've supplied corrupted data with:\n"
                + f"- `n_grains = {self.n_grains}`,\n"
                + f"- `fractions[0].shape = {self.fractions[0].shape}`, and\n"
                + f"- `orientations[0].shape = {self.orientations[0].shape}`."
            )

    def load(self, filename, postfix=None):
        """Load CPO data from a `numpy` NPZ file.

        If `postfix` is not `None`, data is read from fields ending with "`_postfix`".

        See also: `Mineral.save`, `Mineral.from_file`.

        """
        if not filename.endswith(".npz"):
            raise ValueError(
                f"Must only load from numpy NPZ format. Cannot load from {filename}."
            )
        data = np.load(filename)
        if postfix is not None:
            phase, fabric, regime = data[f"meta_{postfix}"]
            self.fractions = list(data[f"fractions_{postfix}"])
            self.orientations = list(data[f"orientations_{postfix}"])
        else:
            phase, fabric, regime = data["meta"]
            self.fractions = list(data["fractions"])
            self.orientations = list(data["orientations"])

        self.phase = phase
        self.fabric = fabric
        self.regime = regime
        self.orientations_init = self.orientations[0]
        self.fractions_init = self.fractions[0]

    @classmethod
    def from_file(cls, filename, postfix=None):
        """Construct a `Mineral` instance using data from a `numpy` NPZ file.

        If `postfix` is not `None`, data is read from fields ending with “`_postfix`”.

        See also: `Mineral.save`, `Mineral.load`.

        """
        if not filename.endswith(".npz"):
            raise ValueError(
                f"Must only load from numpy NPZ format. Cannot load from {filename}."
            )
        data = np.load(filename)
        if postfix is not None:
            phase, fabric, regime = data[f"meta_{postfix}"]
            fractions = list(data[f"fractions_{postfix}"])
            orientations = list(data[f"orientations_{postfix}"])
        else:
            phase, fabric, regime = data["meta"]
            fractions = list(data["fractions"])
            orientations = list(data["orientations"])

        mineral = cls(
            phase,
            fabric,
            regime,
            n_grains=len(fractions[0]),
            fractions_init=fractions[0],
            orientations_init=orientations[0],
        )
        mineral.fractions = fractions
        mineral.orientations = orientations
        return mineral



"""> PyDRex: Tests for tensor operations."""

import numpy as np

from pydrex import tensors as _tensors
from pydrex.minerals import ENSTATITE_STIFFNESS, OLIVINE_STIFFNESS


def test_voigt_decompose():
    """Test decomposition of Voigt 6x6 matrix into distinct contractions."""
    olivine_tensor = _tensors.voigt_to_elastic_tensor(OLIVINE_STIFFNESS)
    dilat, voigt = _tensors.voigt_decompose(OLIVINE_STIFFNESS)
    np.testing.assert_array_equal(np.einsum("ijkk", olivine_tensor), dilat)
    np.testing.assert_array_equal(np.einsum("ijkj", olivine_tensor), voigt)


def test_voigt_tensor():
    """Test elasticity tensor <-> 6x6 Voigt matrix conversions."""
    olivine_tensor = np.array(
        [
            [
                [[320.71, 0.0, 0.0], [0.0, 69.84, 0.0], [0.0, 0.0, 71.22]],
                [[0.0, 78.36, 0.0], [78.36, 0.0, 0.0], [0.0, 0.0, 0.0]],
                [[0.0, 0.0, 77.67], [0.0, 0.0, 0.0], [77.67, 0.0, 0.0]],
            ],
            [
                [[0.0, 78.36, 0.0], [78.36, 0.0, 0.0], [0.0, 0.0, 0.0]],
                [[69.84, 0.0, 0.0], [0.0, 197.25, 0.0], [0.0, 0.0, 74.8]],
                [[0.0, 0.0, 0.0], [0.0, 0.0, 63.77], [0.0, 63.77, 0.0]],
            ],
            [
                [[0.0, 0.0, 77.67], [0.0, 0.0, 0.0], [77.67, 0.0, 0.0]],
                [[0.0, 0.0, 0.0], [0.0, 0.0, 63.77], [0.0, 63.77, 0.0]],
                [[71.22, 0.0, 0.0], [0.0, 74.8, 0.0], [0.0, 0.0, 234.32]],
            ],
        ]
    )
    np.testing.assert_array_equal(
        _tensors.voigt_to_elastic_tensor(OLIVINE_STIFFNESS),
        olivine_tensor,
    )
    np.testing.assert_array_equal(
        _tensors.elastic_tensor_to_voigt(
            _tensors.voigt_to_elastic_tensor(OLIVINE_STIFFNESS)
        ),
        OLIVINE_STIFFNESS,
    )
    np.testing.assert_array_equal(
        _tensors.voigt_to_elastic_tensor(
            _tensors.elastic_tensor_to_voigt(olivine_tensor)
        ),
        olivine_tensor,
    )
    np.testing.assert_array_equal(
        _tensors.elastic_tensor_to_voigt(
            _tensors.voigt_to_elastic_tensor(ENSTATITE_STIFFNESS)
        ),
        ENSTATITE_STIFFNESS,
    )


def test_voigt_to_vector():
    """Test Voigt vector construction."""
    np.testing.assert_allclose(
        _tensors.voigt_matrix_to_vector(ENSTATITE_STIFFNESS),
        np.array(
            [
                236.9,
                180.5,
                230.4,
                80.32733034,
                89.37829714,
                112.57139956,
                168.6,
                158.8,
                160.2,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
            ]
        ),
        atol=1e-9,
    )
    np.testing.assert_array_equal(
        OLIVINE_STIFFNESS,
        _tensors.voigt_vector_to_matrix(
            _tensors.voigt_matrix_to_vector(OLIVINE_STIFFNESS),
        ),
    )
    r = np.random.rand(6, 6)
    np.testing.assert_array_equal(
        _tensors.voigt_matrix_to_vector(r),
        np.array(
            [
                r[0, 0],
                r[1, 1],
                r[2, 2],
                np.sqrt(2) * r[1, 2],
                np.sqrt(2) * r[2, 0],
                np.sqrt(2) * r[0, 1],
                2 * r[3, 3],
                2 * r[4, 4],
                2 * r[5, 5],
                2 * r[0, 3],
                2 * r[1, 4],
                2 * r[2, 5],
                2 * r[2, 3],
                2 * r[0, 4],
                2 * r[1, 5],
                2 * r[1, 3],
                2 * r[2, 4],
                2 * r[0, 5],
                2 * np.sqrt(2) * r[4, 5],
                2 * np.sqrt(2) * r[5, 3],
                2 * np.sqrt(2) * r[3, 4],
            ]
        ),
    )



"""> Configuration and fixtures for PyDRex tests."""

import sys

import matplotlib
import pytest
from _pytest.logging import LoggingPlugin, _LiveLoggingStreamHandler

from pydrex import io as _io
from pydrex import logger as _log
from pydrex import mock as _mock
from pydrex import utils as _utils
from tests import test_vortex_2d as _test_vortex_2d

_log.quiet_aliens()  # Stop imported modules from spamming the logs.
_, HAS_RAY = _utils.import_proc_pool()
if HAS_RAY:
    import ray


# Set up custom pytest CLI arguments.
def pytest_addoption(parser):
    parser.addoption(
        "--outdir",
        metavar="DIR",
        default=None,
        help="output directory in which to store PyDRex figures/logs",
    )
    parser.addoption(
        "--runslow",
        action="store_true",
        default=False,
        help="run slow tests (HPC cluster recommended, large memory requirement)",
    )
    parser.addoption(
        "--runbig",
        action="store_true",
        default=False,
        help="run tests which are fast enough for home PCs but require 16GB RAM",
    )
    parser.addoption(
        "--ncpus",
        default=_utils.default_ncpus(),
        type=int,
        help="number of CPUs to use for tests that support multiprocessing",
    )
    parser.addoption(
        "--fontsize",
        default=None,
        type=int,
        help="set explicit font size for output figures",
    )
    parser.addoption(
        "--markersize",
        default=None,
        type=int,
        help="set explicit marker size for output figures",
    )
    parser.addoption(
        "--linewidth",
        default=None,
        type=int,
        help="set explicit line width for output figures",
    )


# The default pytest logging plugin always creates its own handlers...
class PytestConsoleLogger(LoggingPlugin):
    """Pytest plugin that allows linking up a custom console logger."""

    name = "pytest-console-logger"

    def __init__(self, config, *args, **kwargs):
        super().__init__(config, *args, **kwargs)
        terminal_reporter = config.pluginmanager.get_plugin("terminalreporter")
        capture_manager = config.pluginmanager.get_plugin("capturemanager")
        handler = _LiveLoggingStreamHandler(terminal_reporter, capture_manager)
        handler.setFormatter(_log.CONSOLE_LOGGER.formatter)
        handler.setLevel(_log.CONSOLE_LOGGER.level)
        self.log_cli_handler = handler

    # Override original, which tries to delete some silly globals that we aren't
    # using anymore, this might break the (already quite broken) -s/--capture.
    @pytest.hookimpl(hookwrapper=True)
    def pytest_runtest_teardown(self, item):
        self.log_cli_handler.set_when("teardown")
        yield from self._runtest_for(item, "teardown")


@pytest.hookimpl(trylast=True)
def pytest_configure(config):
    config.addinivalue_line("markers", "slow: mark test as slow to run")
    config.addinivalue_line("markers", "big: mark test as requiring 16GB RAM")

    # Set custom Matplotlib parameters.
    # Alternatively inject a call to `matplotlib.style.use` before starting pytest.
    if config.option.fontsize is not None:
        matplotlib.rcParams["font.size"] = config.option.fontsize
    if config.option.markersize is not None:
        matplotlib.rcParams["lines.markersize"] = config.option.markersize
    if config.option.linewidth is not None:
        matplotlib.rcParams["lines.linewidth"] = config.option.linewidth

    # Hook up our logging plugin last,
    # it relies on terminalreporter and capturemanager.
    if config.option.verbose > 0:
        terminal_reporter = config.pluginmanager.get_plugin("terminalreporter")
        capture_manager = config.pluginmanager.get_plugin("capturemanager")
        handler = _LiveLoggingStreamHandler(terminal_reporter, capture_manager)
        handler.setFormatter(_log.CONSOLE_LOGGER.formatter)
        handler.setLevel(_log.CONSOLE_LOGGER.level)
        _log.LOGGER_PYTEST = handler
        config.pluginmanager.register(
            PytestConsoleLogger(config), PytestConsoleLogger.name
        )


def pytest_collection_modifyitems(config, items):
    if config.getoption("--runslow"):
        # Don't skip slow tests.
        _log.info("running slow tests with %d CPUs", config.getoption("--ncpus"))
    else:
        skip_slow = pytest.mark.skip(reason="need --runslow option to run")
        for item in items:
            if "slow" in item.keywords:
                item.add_marker(skip_slow)

    if config.getoption("--runbig"):
        pass  # Don't skip big tests.
    else:
        skip_big = pytest.mark.skip(reason="need --runbig option to run")
        for item in items:
            if "big" in item.keywords:
                item.add_marker(skip_big)


@pytest.fixture(scope="session")
def outdir(request):
    _outdir = request.config.getoption("--outdir")
    yield _outdir
    #  Create combined ensemble figure for 2D cell tests after they have all finished.
    _test_vortex_2d.TestCellOlivineA._make_ensemble_figure(_outdir)


@pytest.fixture(scope="session")
def ncpus(request):
    return max(1, request.config.getoption("--ncpus"))


@pytest.fixture(scope="session")
def named_tempfile_kwargs(request):
    if sys.platform == "win32":
        return {"delete": False}
    else:
        return dict()


@pytest.fixture(scope="session")
def ray_session():
    if HAS_RAY:
        # NOTE: Expects a running Ray cluster with a number of CPUS matching --ncpus.
        if not ray.is_initialized():
            ray.init(address="auto")
            _log.info("using Ray cluster with %s", ray.cluster_resources())
        yield
        if ray.is_initialized():
            ray.shutdown()
    yield


@pytest.fixture(scope="function")
def console_handler(request):
    if request.config.option.verbose > 0:  # Show console logs if -v/--verbose given.
        return request.config.pluginmanager.get_plugin(
            "pytest-console-logger"
        ).log_cli_handler
    return _log.CONSOLE_LOGGER


@pytest.fixture
def params_Fraters2021():
    return _mock.PARAMS_FRATERS2021


@pytest.fixture
def params_Kaminski2001_fig5_solid():
    return _mock.PARAMS_KAMINSKI2001_FIG5_SOLID


@pytest.fixture
def params_Kaminski2001_fig5_shortdash():
    return _mock.PARAMS_KAMINSKI2001_FIG5_SHORTDASH


@pytest.fixture
def params_Kaminski2001_fig5_longdash():
    return _mock.PARAMS_KAMINSKI2001_FIG5_LONGDASH


@pytest.fixture
def params_Kaminski2004_fig4_triangles():
    return _mock.PARAMS_KAMINSKI2004_FIG4_TRIANGLES


@pytest.fixture
def params_Kaminski2004_fig4_squares():
    return _mock.PARAMS_KAMINSKI2004_FIG4_SQUARES


@pytest.fixture
def params_Kaminski2004_fig4_circles():
    return _mock.PARAMS_KAMINSKI2004_FIG4_CIRCLES


@pytest.fixture
def params_Hedjazian2017():
    return _mock.PARAMS_HEDJAZIAN2017


@pytest.fixture(scope="session", params=[100, 500, 1000, 5000, 10000])
def n_grains(request):
    return request.param


@pytest.fixture(scope="session", params=[[1, 0, 0], [0, 1, 0], [0, 0, 1]])
def hkl(request):
    return request.param


@pytest.fixture(scope="session", params=["xz", "yz", "xy"])
def ref_axes(request):
    return request.param


@pytest.fixture(scope="session")
def seeds():
    """1000 unique seeds for ensemble runs that need an RNG seed."""
    return _io.read_scsv(_io.data("rng") / "seeds.scsv").seeds


@pytest.fixture(scope="session")
def seed():
    """Default seed for test RNG."""
    return 8816


@pytest.fixture(scope="session")
def seeds_nearX45():
    """41 seeds which have the initial hexagonal symmetry axis near 45° from X."""
    return _io.read_scsv(_io.data("rng") / "hexaxis_nearX45_seeds.scsv").seeds



"""> PyDRex: Statistical methods for orientation and elasticity data."""

import itertools as it

import numpy as np
import scipy.special as sp
from scipy.spatial.transform import Rotation

from pydrex import geometry as _geo
from pydrex import stats as _stats
from pydrex import utils as _utils


def resample_orientations(orientations, fractions, n_samples=None, seed=None):
    """Return new samples from `orientations` weighted by the volume distribution.

    Args:
    - `orientations` (array) — NxMx3x3 array of orientations
    - `fractions` (array) — NxM array of grain volume fractions
    - `n_samples` (int) — optional number of samples to return, default is M
    - `seed` (int) — optional seed for the random number generator, which is used to
      pick random grain volume samples from the discrete distribution

    Returns the Nx`n_samples`x3x3 orientations and associated sorted (ascending) grain
    volumes.

    """
    # Allow lists of Rotation.as_matrix() inputs.
    _orientations = np.asarray(orientations)
    _fractions = np.asarray(fractions)
    # Fail early to prevent possibly expensive data processing on incorrect data arrays.
    if (
        len(_orientations.shape) != 4
        or len(_fractions.shape) != 2
        or _orientations.shape[0] != _fractions.shape[0]
        or _orientations.shape[1] != _fractions.shape[1]
        or _orientations.shape[2] != _orientations.shape[3] != 3
    ):
        raise ValueError(
            "invalid shape of input arrays,"
            + f" got orientations of shape {_orientations.shape}"
            + f" and fractions of shape {_fractions.shape}"
        )
    rng = np.random.default_rng(seed=seed)
    if n_samples is None:
        n_samples = _fractions.shape[1]
    out_orientations = np.empty((len(_fractions), n_samples, 3, 3))
    out_fractions = np.empty((len(_fractions), n_samples))
    for i, (frac, orient) in enumerate(zip(_fractions, _orientations, strict=True)):
        sort_ascending = np.argsort(frac)
        # Cumulative volume fractions.
        frac_ascending = frac[sort_ascending]
        cumfrac = frac_ascending.cumsum()
        # Force cumfrac[-1] to be equal to sum(frac_ascending) i.e. 1.
        cumfrac[-1] = 1.0
        # Number of new samples with volume less than each cumulative fraction.
        count_less = np.searchsorted(cumfrac, rng.random(n_samples))
        out_orientations[i, ...] = orient[sort_ascending][count_less]
        out_fractions[i, ...] = frac_ascending[count_less]
    return out_orientations, out_fractions


def _scatter_matrix(orientations, row):
    # Lower triangular part of the symmetric scatter (inertia) matrix,
    # see eq. 2.4 in Watson 1966 or eq. 9.2.10 in Mardia & Jupp 2009 (with n = 1),
    # it's a summation of the outer product of the [h, k, l] vector with itself,
    # so taking the row assumes that `orientations` are passive rotations of the
    # reference frame [h, k, l] vector.
    scatter = np.zeros((3, 3))
    scatter[0, 0] = np.sum(orientations[:, row, 0] ** 2)
    scatter[1, 1] = np.sum(orientations[:, row, 1] ** 2)
    scatter[2, 2] = np.sum(orientations[:, row, 2] ** 2)
    scatter[1, 0] = np.sum(orientations[:, row, 0] * orientations[:, row, 1])
    scatter[2, 0] = np.sum(orientations[:, row, 0] * orientations[:, row, 2])
    scatter[2, 1] = np.sum(orientations[:, row, 1] * orientations[:, row, 2])
    return scatter


def misorientation_hist(orientations, system: _geo.LatticeSystem, bins=None):
    r"""Calculate misorientation histogram for polycrystal orientations.

    The `bins` argument is passed to `numpy.histogram`.
    If left as `None`, 1° bins will be used as recommended by the reference paper.
    The `symmetry` argument specifies the lattice system which determines intrinsic
    symmetry degeneracies and the maximum allowable misorientation angle.
    See `_geo.LatticeSystem` for supported systems.

    .. warning::
        This method must be able to allocate $ \frac{N!}{(N-2)!} × 4M $ floats
        for N the length of `orientations` and M the number of symmetry operations for
        the given `system` (`numpy.float32` values are used to reduce the memory
        requirements)

    See [Skemer et al. (2005)](https://doi.org/10.1016/j.tecto.2005.08.023).

    """
    symmetry_ops = _geo.symmetry_operations(system)
    # Compute and bin misorientation angles from orientation data.
    q1_array = np.empty(
        (sp.comb(len(orientations), 2, exact=True), len(symmetry_ops), 4),
        dtype=np.float32,
    )
    q2_array = np.empty(
        (sp.comb(len(orientations), 2, exact=True), len(symmetry_ops), 4),
        dtype=np.float32,
    )
    for i, e in enumerate(  # Copy is required for proper object referencing in Ray.
        it.combinations(Rotation.from_matrix(orientations.copy()).as_quat(), 2)
    ):
        q1, q2 = list(e)
        for j, qs in enumerate(symmetry_ops):
            if qs.shape == (4, 4):  # Reflection, not a proper rotation.
                q1_array[i, j] = qs @ q1
                q2_array[i, j] = qs @ q2
            else:
                q1_array[i, j] = _utils.quat_product(qs, q1)
                q2_array[i, j] = _utils.quat_product(qs, q2)

    misorientations_data = _geo.misorientation_angles(q1_array, q2_array)
    θmax = _stats._max_misorientation(system)
    return np.histogram(misorientations_data, bins=θmax, range=(0, θmax), density=True)


def misorientations_random(low, high, system: _geo.LatticeSystem):
    """Get expected count of misorientation angles for an isotropic aggregate.

    Estimate the expected number of misorientation angles between grains
    that would fall within $($`low`, `high`$)$ in degrees for an aggregate
    with randomly oriented grains, where `low` $∈ [0, $`high`$)$,
    and `high` is bounded by the maximum theoretical misorientation angle
    for the given lattice symmetry system.
    See `_geo.LatticeSystem` for supported systems.

    """
    # TODO: Add cubic system: [Handscomb 1958](https://doi.org/10.4153/CJM-1958-010-0)
    max_θ = _max_misorientation(system)
    M, N = system.value
    if not 0 <= low <= high <= max_θ:
        raise ValueError(
            f"bounds must obey `low` ∈ [0, `high`) and `high` < {max_θ}.\n"
            + f"You've supplied (`low`, `high`) = ({low}, {high})."
        )

    counts_low = 0  # Number of counts at the lower bin edge.
    counts_high = 0  # ... at the higher bin edge.
    counts_both = [counts_low, counts_high]

    # Some constant factors.
    a = np.tan(np.deg2rad(90 / M))
    b = 2 * np.rad2deg(np.arctan(np.sqrt(1 + a**2)))
    c = round(2 * np.rad2deg(np.arctan(np.sqrt(1 + 2 * a**2))))

    for i, edgeval in enumerate([low, high]):
        d = np.deg2rad(edgeval)

        if 0 <= edgeval <= (180 / M):
            counts_both[i] += (N / 180) * (1 - np.cos(d))

        elif (180 / M) <= edgeval <= (180 * M / N):
            counts_both[i] += (N / 180) * a * np.sin(d)

        elif 90 <= edgeval <= b:
            counts_both[i] += (M / 90) * ((M + a) * np.sin(d) - M * (1 - np.cos(d)))

        elif b <= edgeval <= c:
            ν = np.tan(np.deg2rad(edgeval / 2)) ** 2

            counts_both[i] = (M / 90) * (
                (M + a) * np.sin(d)
                - M * (1 - np.cos(d))
                + (M / 180)
                * (
                    (1 - np.cos(d))
                    * (
                        np.rad2deg(
                            np.arccos((1 - ν * np.cos(np.deg2rad(180 / M))) / (ν - 1))
                        )
                        + 2
                        * np.rad2deg(
                            np.arccos(a / (np.sqrt(ν - a**2) * np.sqrt(ν - 1)))
                        )
                    )
                    - 2
                    * np.sin(d)
                    * (
                        2 * np.rad2deg(np.arccos(a / np.sqrt(ν - 1)))
                        + a * np.rad2deg(np.arccos(1 / np.sqrt(ν - a**2)))
                    )
                )
            )
        else:
            assert False  # Should never happen.

    return np.sum(counts_both) / 2


def _max_misorientation(system):
    # Maximum misorientation angle for two grains of the given lattice symmetry system.
    s = _geo.LatticeSystem
    match system:
        case s.orthorhombic | s.rhombohedral:
            max_θ = 120
        case s.tetragonal | s.hexagonal:
            max_θ = 90
        case s.triclinic | s.monoclinic:
            max_θ = 180
        case _:
            raise ValueError(f"unsupported lattice system: {system}")
    return max_θ


def point_density(
    x_data,
    y_data,
    z_data,
    gridsteps=101,
    weights=1,
    kernel="linear_inverse_kamb",
    axial=True,
    **kwargs,
):
    """Estimate point density of orientation data on the unit sphere.

    Estimates the density of orientations on the unit sphere by counting the input data
    that falls within small areas around a uniform grid of spherical counting locations.
    The input data is expected in cartesian coordinates, and the contouring is performed
    using kernel functions defined in [Vollmer 1995](https://doi.org/10.1016/0098-3004(94)00058-3).
    The following optional parameters control the contouring method:
    - `gridsteps` (int) — the number of steps, i.e. number of points along a diameter of
        the spherical counting grid
    - `weights` (array) — auxiliary weights for each data point
    - `kernel` (string) — the name of the kernel function to use, see
      `SPHERICAL_COUNTING_KERNELS`
    - `axial` (bool) — toggle axial versions of the kernel functions
        (for crystallographic data this should normally be kept as `True`)

    Any other keyword arguments are passed to the kernel function calls.
    Most kernels accept a parameter `σ` to control the degree of smoothing.

    """
    if kernel not in SPHERICAL_COUNTING_KERNELS:
        raise ValueError(f"kernel '{kernel}' is not supported")
    weights = np.asarray(weights, dtype=np.float64)

    # Create a grid of counters on a cylinder.
    ρ_grid, h_grid = np.mgrid[-np.pi : np.pi : gridsteps * 1j, -1 : 1 : gridsteps * 1j]
    # Project onto the sphere using the equal-area projection with centre at (0, 0).
    λ_grid = ρ_grid
    ϕ_grid = np.arcsin(h_grid)
    x_counters, y_counters, z_counters = _geo.to_cartesian(
        np.pi / 2 - λ_grid.ravel(), np.pi / 2 - ϕ_grid.ravel()
    )

    # Basically, we can't model this as a convolution as we're not in Euclidean space,
    # so we have to iterate through and call the kernel function at each "counter".
    data = np.column_stack([x_data, y_data, z_data])
    counters = np.column_stack([x_counters, y_counters, z_counters])
    totals = np.empty(counters.shape[0])
    for i, counter in enumerate(counters):
        products = np.dot(data, counter)
        if axial:
            products = np.abs(products)
        density, scale = SPHERICAL_COUNTING_KERNELS[kernel](
            products, axial=axial, **kwargs
        )
        density *= weights
        totals[i] = (density.sum() - 0.5) / scale

    X_counters, Y_counters = _geo.lambert_equal_area(x_counters, y_counters, z_counters)

    # Normalise to mean, which estimates the density for a "uniform" distribution.
    totals /= totals.mean()
    totals[totals < 0] = 0
    # print(totals.min(), totals.mean(), totals.max())
    return (
        np.reshape(X_counters, ρ_grid.shape),
        np.reshape(Y_counters, ρ_grid.shape),
        np.reshape(totals, ρ_grid.shape),
    )


def _kamb_radius(n, σ, axial):
    """Radius of kernel for Kamb-style smoothing."""
    r = σ**2 / (float(n) + σ**2)
    if axial is True:
        return 1 - r
    return 1 - 2 * r


def _kamb_units(n, radius):
    """Normalization function for Kamb-style counting."""
    return np.sqrt(n * radius * (1 - radius))


def exponential_kamb(cos_dist, σ=10, axial=True):
    """Kernel function from Vollmer 1995 for exponential smoothing."""
    n = float(cos_dist.size)
    if axial:
        f = 2 * (1.0 + n / σ**2)
        units = np.sqrt(n * (f / 2.0 - 1) / f**2)
    else:
        f = 1 + n / σ**2
        units = np.sqrt(n * (f - 1) / (4 * f**2))

    count = np.exp(f * (cos_dist - 1))
    return count, units


def linear_inverse_kamb(cos_dist, σ=10, axial=True):
    """Kernel function from Vollmer 1995 for linear smoothing."""
    n = float(cos_dist.size)
    radius = _kamb_radius(n, σ, axial=axial)
    f = 2 / (1 - radius)
    cos_dist = cos_dist[cos_dist >= radius]
    count = f * (cos_dist - radius)
    return count, _kamb_units(n, radius)


def square_inverse_kamb(cos_dist, σ=10, axial=True):
    """Kernel function from Vollmer 1995 for inverse square smoothing."""
    n = float(cos_dist.size)
    radius = _kamb_radius(n, σ, axial=axial)
    f = 3 / (1 - radius) ** 2
    cos_dist = cos_dist[cos_dist >= radius]
    count = f * (cos_dist - radius) ** 2
    return count, _kamb_units(n, radius)


def kamb_count(cos_dist, σ=10, axial=True):
    """Original Kamb 1959 kernel function (raw count within radius)."""
    n = float(cos_dist.size)
    dist = _kamb_radius(n, σ, axial=axial)
    count = (cos_dist >= dist).astype(float)
    return count, _kamb_units(n, dist)


def schmidt_count(cos_dist, axial=None):
    """Schmidt (a.k.a. 1%) counting kernel function."""
    radius = 0.01
    count = ((1 - cos_dist) <= radius).astype(float)
    # To offset the count.sum() - 0.5 required for the kamb methods...
    count = 0.5 / count.size + count
    return count, (cos_dist.size * radius)


SPHERICAL_COUNTING_KERNELS = {
    "kamb_count": kamb_count,
    "schmidt_count": schmidt_count,
    "exponential_kamb": exponential_kamb,
    "linear_inverse_kamb": linear_inverse_kamb,
    "square_inverse_kamb": square_inverse_kamb,
}
"""Kernel functions that return an un-summed distribution and a normalization factor.

Supported kernel functions are based on the discussion in
[Vollmer 1995](https://doi.org/10.1016/0098-3004(94)00058-3).
Kamb methods accept the parameter `σ` (default: 10) to control the degree of smoothing.
Values lower than 3 and higher than 20 are not recommended.

"""



"""> PyDRex: Tests for geometric conversions and projections."""

import numpy as np

from pydrex import geometry as _geo
from pydrex import io as _io


def test_poles_example(hkl, ref_axes):
    """Test poles (directions of crystallographic axes) of example data."""
    ref_data = np.load(
        _io.resolve_path(
            _io.data("outputs")
            / f"example_CPO_poles_{_io.stringify(hkl)}{ref_axes}.npz"
        )
    )
    resampled_data = np.load(
        _io.resolve_path(_io.data("outputs") / "example_CPO_resampled.npz")
    )
    xvals, yvals, zvals = _geo.poles(
        resampled_data["orientations"],
        hkl=hkl,
        ref_axes=ref_axes,
    )
    np.testing.assert_allclose(ref_data["xvals"], xvals, atol=1e-16, rtol=0)
    np.testing.assert_allclose(ref_data["yvals"], yvals, atol=1e-16, rtol=0)
    np.testing.assert_allclose(ref_data["zvals"], zvals, atol=1e-16, rtol=0)


def test_lambert_equal_area(seed):
    """Test Lambert equal area projection."""
    x, y = np.mgrid[-1:1:11j, -1:1:11j]
    x_flat = [j for i in x for j in i]
    y_flat = [j for i in y for j in i]
    # Uniform samples on the unit disk, this is tested in the Shirley doctest example.
    x_disk, y_disk = _geo.shirley_concentric_squaredisk(x_flat, y_flat)
    # Project onto the unit sphere by adding z = ± (1 - r).
    # Then project back onto the disk using Lambert equal-area, should be the same.
    rng = np.random.default_rng(seed=seed)
    sign = rng.integers(low=0, high=2, size=len(x_disk))
    x_laea, y_laea = _geo.lambert_equal_area(
        x_disk, y_disk, (-1) ** sign * (1 - (x_disk**2 + y_disk**2))
    )
    np.testing.assert_allclose(x_disk, x_laea, atol=1e-15, rtol=0)
    np.testing.assert_allclose(y_disk, y_laea, atol=1e-15, rtol=0)
